import importlib.metadata

from .array import (
    bin,
    blur,
    blur2,
    blur_image,
    difference,
    disjoint,
    extent,
    group,
    groups,
    index,
    indexes,
    intersection,
    nice,
    rollup,
    rollups,
    subset,
    superset,
    threshold_freedman_diaconis,
    threshold_scott,
    threshold_sturges,
    tick_increment,
    tick_step,
    ticks,
    union,
)
from .axis import axis_bottom, axis_left, axis_right, axis_top
from .chord import (
    chord,
    chord_directed,
    chord_transpose,
    ribbon,
    ribbon_arrow,
)
from .color import (
    color,
    cubehelix,
    gray,
    hcl,
    hsl,
    lab,
    lch,
    rgb,
)
from .contour import Contours as contours
from .contour import Density as contour_density
from .delaunay import Delaunay, Voronoi
from .ease import (
    ease_back,
    ease_back_in,
    ease_back_in_out,
    ease_back_out,
    ease_bounce,
    ease_bounce_in,
    ease_bounce_in_out,
    ease_bounce_out,
    ease_circle,
    ease_circle_in,
    ease_circle_in_out,
    ease_circle_out,
    ease_cubic,
    ease_cubic_in,
    ease_cubic_in_out,
    ease_cubic_out,
    ease_elastic,
    ease_elastic_in,
    ease_elastic_in_out,
    ease_elastic_out,
    ease_exp,
    ease_exp_in,
    ease_exp_in_out,
    ease_exp_out,
    ease_linear,
    ease_poly,
    ease_poly_in,
    ease_poly_in_out,
    ease_poly_out,
    ease_quad,
    ease_quad_in,
    ease_quad_in_out,
    ease_quad_out,
    ease_sin,
    ease_sin_in,
    ease_sin_in_out,
    ease_sin_out,
)
from .force import (
    force_center,
    force_collide,
    force_link,
    force_many_body,
    force_radial,
    force_simulation,
    force_x,
    force_y,
)
from .format import (
    format_default_locale,
    format_prefix,
    precision_fixed,
    precision_prefix,
    precision_round,
)
from .format import (
    locale_format as format,
)
from .geo import (
    geo_albers,
    geo_albers_usa,
    geo_area,
    geo_azimuthal_equal_area,
    geo_azimuthal_equidistant,
    geo_bounds,
    geo_centroid,
    geo_circle,
    geo_clip_antimeridian,
    geo_clip_circle,
    geo_clip_rectangle,
    geo_conic_conformal,
    geo_conic_equal_area,
    geo_conic_equidistant,
    geo_contains,
    geo_distance,
    geo_equal_earth,
    geo_equirectangular,
    geo_gnomonic,
    geo_graticule,
    geo_graticule_10,
    geo_identity,
    geo_interpolate,
    geo_length,
    geo_mercator,
    geo_natural_earth_1,
    geo_orthographic,
    geo_path,
    geo_projection,
    geo_projection_mutator,
    geo_rotation,
    geo_stereographic,
    geo_stream,
    geo_transform,
    geo_transverse_mercator,
)
from .hierarchy import (
    cluster,
    hierarchy,
    pack,
    pack_enclose,
    pack_siblings,
    partition,
    stratify,
    tree,
    treemap,
    treemap_binary,
    treemap_dice,
    treemap_resquarify,
    treemap_slice,
    treemap_slice_dice,
    treemap_squarify,
)
from .interpolate import (
    interpolate,
    interpolate_array,
    interpolate_basis,
    interpolate_basis_closed,
    interpolate_cubehelix,
    interpolate_cubehelix_long,
    interpolate_date,
    interpolate_discrete,
    interpolate_hcl,
    interpolate_hcl_long,
    interpolate_hsl,
    interpolate_hsl_long,
    interpolate_hue,
    interpolate_lab,
    interpolate_number,
    interpolate_number_array,
    interpolate_object,
    interpolate_rgb,
    interpolate_rgb_basis,
    interpolate_rgb_basis_closed,
    interpolate_round,
    interpolate_string,
    interpolate_zoom,
    piecewise,
    quantize,
)
from .path import Path as path
from .polygon import (
    polygon_area,
    polygon_centroid,
    polygon_contains,
    polygon_hull,
    polygon_length,
)
from .quadtree import quadtree
from .scale import (
    scale_band,
    scale_diverging,
    scale_diverging_log,
    scale_diverging_pow,
    scale_diverging_sqrt,
    scale_diverging_symlog,
    scale_identity,
    scale_linear,
    scale_log,
    scale_ordinal,
    scale_point,
    scale_pow,
    scale_quantile,
    scale_quantize,
    scale_radial,
    scale_sequential,
    scale_sequential_log,
    scale_sequential_pow,
    scale_sequential_quantile,
    scale_sequential_sqrt,
    scale_sequential_symlog,
    scale_sqrt,
    scale_symlog,
    scale_threshold,
    scale_time,
    tick_format,
)
from .scale_chromatic import (
    SCHEME_ACCENT,
    SCHEME_BLUES,
    SCHEME_BRBG,
    SCHEME_BUGN,
    SCHEME_BUPU,
    SCHEME_CATEGORY_10,
    SCHEME_DARK_2,
    SCHEME_GNBU,
    SCHEME_GREENS,
    SCHEME_GREYS,
    SCHEME_OBSERVABLE_10,
    SCHEME_ORANGES,
    SCHEME_ORRD,
    SCHEME_PAIRED,
    SCHEME_PASTEL_1,
    SCHEME_PASTEL_2,
    SCHEME_PIYG,
    SCHEME_PRGN,
    SCHEME_PUBU,
    SCHEME_PUBUGN,
    SCHEME_PUOR,
    SCHEME_PURD,
    SCHEME_PURPLES,
    SCHEME_RDBU,
    SCHEME_RDGY,
    SCHEME_RDPU,
    SCHEME_RDYLBU,
    SCHEME_RDYLGN,
    SCHEME_REDS,
    SCHEME_SET_1,
    SCHEME_SET_2,
    SCHEME_SET_3,
    SCHEME_SPECTRAL,
    SCHEME_TABLEAU_10,
    SCHEME_YLGN,
    SCHEME_YLGNBU,
    SCHEME_YLORBR,
    SCHEME_YLORRD,
    interpolate_blues,
    interpolate_brbg,
    interpolate_bugn,
    interpolate_bupu,
    interpolate_cividis,
    interpolate_cool,
    interpolate_cubehelix_default,
    interpolate_gnbu,
    interpolate_greens,
    interpolate_greys,
    interpolate_inferno,
    interpolate_magma,
    interpolate_oranges,
    interpolate_orrd,
    interpolate_piyg,
    interpolate_plasma,
    interpolate_prgn,
    interpolate_pubu,
    interpolate_pubugn,
    interpolate_puor,
    interpolate_purd,
    interpolate_purples,
    interpolate_rainbow,
    interpolate_rdbu,
    interpolate_rdgy,
    interpolate_rdpu,
    interpolate_rdylbu,
    interpolate_rdylgn,
    interpolate_reds,
    interpolate_sinebow,
    interpolate_spectral,
    interpolate_turbo,
    interpolate_viridis,
    interpolate_warm,
    interpolate_ylgn,
    interpolate_ylgnbu,
    interpolate_ylorbr,
    interpolate_ylorrd,
)
from .selection import create, select
from .shape import (
    SYMBOLS_FILL,
    SYMBOLS_STROKE,
    area_radial,
    curve_basis,
    curve_basis_closed,
    curve_basis_open,
    curve_bump_radial,
    curve_bump_x,
    curve_bump_y,
    curve_bundle,
    curve_cardinal,
    curve_cardinal_closed,
    curve_cardinal_open,
    curve_catmull_rom,
    curve_catmull_rom_closed,
    curve_catmull_rom_open,
    curve_linear,
    curve_linear_closed,
    curve_monotone_x,
    curve_monotone_y,
    curve_natural,
    curve_step,
    curve_step_after,
    curve_step_before,
    line_radial,
    link_horizontal,
    link_radial,
    link_vertical,
    point_radial,
    symbol_asterisk,
    symbol_circle,
    symbol_cross,
    symbol_diamond,
    symbol_diamond2,
    symbol_plus,
    symbol_square,
    symbol_square2,
    symbol_star,
    symbol_times,
    symbol_triangle,
    symbol_triangle2,
    symbol_wye,
)
from .shape import (
    Arc as arc,
)
from .shape import (
    Area as area,
)
from .shape import (
    Line as line,
)
from .shape import (
    Link as link,
)
from .shape import (
    Pie as pie,
)
from .shape import (
    Stack as stack,
)
from .shape import (
    Symbol as symbol,
)
from .shape import (
    offset_diverging as stack_offset_diverging,
)
from .shape import (
    offset_expand as stack_offset_expand,
)
from .shape import (
    offset_none as stack_offset_none,
)
from .shape import (
    offset_silhouette as stack_offset_silhouette,
)
from .shape import (
    offset_wiggle as stack_offset_wiggle,
)
from .shape import (
    order_appearance as stack_order_appearance,
)
from .shape import (
    order_ascending as stack_order_ascending,
)
from .shape import (
    order_descending as stack_order_descending,
)
from .shape import (
    order_inside_out as stack_order_inside_out,
)
from .shape import (
    order_none as stack_order_none,
)
from .shape import (
    order_reverse as stack_order_reverse,
)
from .time import (
    time_day,
    time_friday,
    time_hour,
    time_millisecond,
    time_minute,
    time_monday,
    time_month,
    time_saturday,
    time_second,
    time_sunday,
    time_thursday,
    time_ticks,
    time_tuesday,
    time_wednesday,
    time_week,
    time_year,
)
from .time_format import (
    iso_format,
    iso_parse,
    time_format,
    time_format_locale,
    time_parse,
)

__version__ = importlib.metadata.version(__package__)

__all__ = [
    "Delaunay",
    "SCHEME_ACCENT",
    "SCHEME_BLUES",
    "SCHEME_BRBG",
    "SCHEME_BUGN",
    "SCHEME_BUPU",
    "SCHEME_CATEGORY_10",
    "SCHEME_DARK_2",
    "SCHEME_GNBU",
    "SCHEME_GREENS",
    "SCHEME_GREYS",
    "SCHEME_OBSERVABLE_10",
    "SCHEME_ORANGES",
    "SCHEME_ORRD",
    "SCHEME_PAIRED",
    "SCHEME_PASTEL_1",
    "SCHEME_PASTEL_2",
    "SCHEME_PIYG",
    "SCHEME_PRGN",
    "SCHEME_PUBU",
    "SCHEME_PUBUGN",
    "SCHEME_PUOR",
    "SCHEME_PURD",
    "SCHEME_PURPLES",
    "SCHEME_RDBU",
    "SCHEME_RDGY",
    "SCHEME_RDPU",
    "SCHEME_RDYLBU",
    "SCHEME_RDYLGN",
    "SCHEME_REDS",
    "SCHEME_SET_1",
    "SCHEME_SET_2",
    "SCHEME_SET_3",
    "SCHEME_SPECTRAL",
    "SCHEME_TABLEAU_10",
    "SCHEME_YLGN",
    "SCHEME_YLGNBU",
    "SCHEME_YLORBR",
    "SCHEME_YLORRD",
    "SYMBOLS_FILL",
    "SYMBOLS_STROKE",
    "Voronoi",
    "arc",
    "area",
    "area_radial",
    "axis_bottom",
    "axis_left",
    "axis_right",
    "axis_top",
    "bin",
    "blur",
    "blur2",
    "blur_image",
    "chord",
    "chord_directed",
    "chord_transpose",
    "cluster",
    "color",
    "contour_density",
    "contours",
    "create",
    "cubehelix",
    "curve_basis",
    "curve_basis_closed",
    "curve_basis_open",
    "curve_bump_radial",
    "curve_bump_x",
    "curve_bump_y",
    "curve_bundle",
    "curve_cardinal",
    "curve_cardinal_closed",
    "curve_cardinal_open",
    "curve_catmull_rom",
    "curve_catmull_rom_closed",
    "curve_catmull_rom_open",
    "curve_linear",
    "curve_linear_closed",
    "curve_monotone_x",
    "curve_monotone_y",
    "curve_natural",
    "curve_step",
    "curve_step_after",
    "curve_step_before",
    "difference",
    "disjoint",
    "ease_back",
    "ease_back_in",
    "ease_back_in_out",
    "ease_back_out",
    "ease_bounce",
    "ease_bounce_in",
    "ease_bounce_in_out",
    "ease_bounce_out",
    "ease_circle",
    "ease_circle_in",
    "ease_circle_in_out",
    "ease_circle_out",
    "ease_cubic",
    "ease_cubic_in",
    "ease_cubic_in_out",
    "ease_cubic_out",
    "ease_elastic",
    "ease_elastic_in",
    "ease_elastic_in_out",
    "ease_elastic_out",
    "ease_exp",
    "ease_exp_in",
    "ease_exp_in_out",
    "ease_exp_out",
    "ease_linear",
    "ease_poly",
    "ease_poly_in",
    "ease_poly_in_out",
    "ease_poly_out",
    "ease_quad",
    "ease_quad_in",
    "ease_quad_in_out",
    "ease_quad_out",
    "ease_sin",
    "ease_sin_in",
    "ease_sin_in_out",
    "ease_sin_out",
    "extent",
    "force_center",
    "force_collide",
    "force_link",
    "force_many_body",
    "force_radial",
    "force_simulation",
    "force_x",
    "force_y",
    "format",
    "format_default_locale",
    "format_prefix",
    "geo_albers",
    "geo_albers_usa",
    "geo_area",
    "geo_azimuthal_equal_area",
    "geo_azimuthal_equidistant",
    "geo_bounds",
    "geo_centroid",
    "geo_circle",
    "geo_clip_antimeridian",
    "geo_clip_circle",
    "geo_clip_rectangle",
    "geo_conic_conformal",
    "geo_conic_equal_area",
    "geo_conic_equidistant",
    "geo_contains",
    "geo_distance",
    "geo_equal_earth",
    "geo_equirectangular",
    "geo_gnomonic",
    "geo_graticule",
    "geo_graticule_10",
    "geo_identity",
    "geo_interpolate",
    "geo_length",
    "geo_mercator",
    "geo_natural_earth_1",
    "geo_orthographic",
    "geo_path",
    "geo_projection",
    "geo_projection_mutator",
    "geo_rotation",
    "geo_stereographic",
    "geo_stream",
    "geo_transform",
    "geo_transverse_mercator",
    "gray",
    "group",
    "groups",
    "hcl",
    "hierarchy",
    "hsl",
    "index",
    "indexes",
    "interpolate",
    "interpolate_array",
    "interpolate_basis",
    "interpolate_basis_closed",
    "interpolate_blues",
    "interpolate_brbg",
    "interpolate_bugn",
    "interpolate_bupu",
    "interpolate_cividis",
    "interpolate_cool",
    "interpolate_cubehelix",
    "interpolate_cubehelix_default",
    "interpolate_cubehelix_long",
    "interpolate_date",
    "interpolate_discrete",
    "interpolate_gnbu",
    "interpolate_greens",
    "interpolate_greys",
    "interpolate_hcl",
    "interpolate_hcl_long",
    "interpolate_hsl",
    "interpolate_hsl_long",
    "interpolate_hue",
    "interpolate_inferno",
    "interpolate_lab",
    "interpolate_magma",
    "interpolate_number",
    "interpolate_number_array",
    "interpolate_object",
    "interpolate_oranges",
    "interpolate_orrd",
    "interpolate_piyg",
    "interpolate_plasma",
    "interpolate_prgn",
    "interpolate_pubu",
    "interpolate_pubugn",
    "interpolate_puor",
    "interpolate_purd",
    "interpolate_purples",
    "interpolate_rainbow",
    "interpolate_rdbu",
    "interpolate_rdgy",
    "interpolate_rdpu",
    "interpolate_rdylbu",
    "interpolate_rdylgn",
    "interpolate_reds",
    "interpolate_rgb",
    "interpolate_rgb_basis",
    "interpolate_rgb_basis_closed",
    "interpolate_round",
    "interpolate_sinebow",
    "interpolate_spectral",
    "interpolate_string",
    "interpolate_turbo",
    "interpolate_viridis",
    "interpolate_warm",
    "interpolate_ylgn",
    "interpolate_ylgnbu",
    "interpolate_ylorbr",
    "interpolate_ylorrd",
    "interpolate_zoom",
    "intersection",
    "iso_format",
    "iso_parse",
    "lab",
    "lch",
    "line",
    "line_radial",
    "line_radial",
    "link",
    "link_horizontal",
    "link_radial",
    "link_vertical",
    "nice",
    "pack",
    "pack_enclose",
    "pack_siblings",
    "partition",
    "path",
    "pie",
    "piecewise",
    "point_radial",
    "polygon_area",
    "polygon_centroid",
    "polygon_contains",
    "polygon_hull",
    "polygon_length",
    "precision_fixed",
    "precision_prefix",
    "precision_round",
    "quadtree",
    "quantize",
    "rgb",
    "ribbon",
    "ribbon_arrow",
    "rollup",
    "rollups",
    "scale_band",
    "scale_diverging",
    "scale_diverging_log",
    "scale_diverging_pow",
    "scale_diverging_sqrt",
    "scale_diverging_symlog",
    "scale_identity",
    "scale_linear",
    "scale_log",
    "scale_ordinal",
    "scale_point",
    "scale_pow",
    "scale_quantile",
    "scale_quantize",
    "scale_radial",
    "scale_sequential",
    "scale_sequential_log",
    "scale_sequential_pow",
    "scale_sequential_quantile",
    "scale_sequential_sqrt",
    "scale_sequential_symlog",
    "scale_sqrt",
    "scale_symlog",
    "scale_threshold",
    "scale_time",
    "select",
    "stack",
    "stack_offset_diverging",
    "stack_offset_expand",
    "stack_offset_none",
    "stack_offset_silhouette",
    "stack_offset_wiggle",
    "stack_order_appearance",
    "stack_order_ascending",
    "stack_order_descending",
    "stack_order_inside_out",
    "stack_order_none",
    "stack_order_reverse",
    "stratify",
    "subset",
    "superset",
    "symbol",
    "symbol_asterisk",
    "symbol_circle",
    "symbol_cross",
    "symbol_diamond",
    "symbol_diamond2",
    "symbol_plus",
    "symbol_square",
    "symbol_square2",
    "symbol_star",
    "symbol_times",
    "symbol_triangle",
    "symbol_triangle2",
    "symbol_wye",
    "threshold_freedman_diaconis",
    "threshold_scott",
    "threshold_sturges",
    "tick_format",
    "tick_increment",
    "tick_step",
    "ticks",
    "time_day",
    "time_format",
    "time_format_locale",
    "time_friday",
    "time_hour",
    "time_millisecond",
    "time_minute",
    "time_monday",
    "time_month",
    "time_parse",
    "time_saturday",
    "time_second",
    "time_sunday",
    "time_thursday",
    "time_ticks",
    "time_tuesday",
    "time_wednesday",
    "time_week",
    "time_year",
    "tree",
    "treemap",
    "treemap_binary",
    "treemap_dice",
    "treemap_resquarify",
    "treemap_slice",
    "treemap_slice_dice",
    "treemap_squarify",
    "union",
]



from ...types import T
from ..series import Series


def order_none(series: list[Series[T]]) -> list[int]:
    """
    Returns the given series order :math:`[0, 1, ..., n - 1]` where :math:`n`
    is the number of elements in series. Thus the stack is given by the key
    accessor.

    Parameters
    ----------
    series : list[Series[T]]
        List of series

    Returns
    -------
    list[int]
        List of ordered indices
    """
    return list(range(len(series)))



from ..common import LineStream


class ClipBuffer(LineStream):
    def __init__(self):
        self._lines = []
        self._line = None

    def line_start(self):
        self._line = []
        self._lines.append(self._line)

    def line_end(self):
        return

    def point(self, x: float, y: float, m: float | None = None):
        self._line.append([x, y, m])

    def rejoin(self):
        if len(self._lines) > 1:
            self._lines.append(self._lines.pop() + self._lines.pop(0))

    def result(self) -> float:
        result = self._lines
        self._lines = []
        self._line = None
        return result

    def __str__(self) -> str:
        return f"ClipBuffer({self._lines})"



from math import isnan

from ...types import T
from ..series import Series
from .none import order_none


def order_ascending(series: list[Series[T]]) -> list[int]:
    """
    Returns a series order such that the smallest series (according to the sum
    of values) is at the bottom.

    Parameters
    ----------
    series : list[Series[T]]
        List of series

    Returns
    -------
    list[int]
        List of ordered indices
    """
    sums = list(map(_sum, series))
    return sorted(order_none(series), key=lambda i: sums[i])


def _sum(series: list[Series[T]]) -> float:
    s = 0
    for i in range(len(series)):
        v = float(series[i][1])
        if not isnan(v):
            s += v
    return s



from __future__ import annotations

from collections.abc import Callable

from ..accessors import required
from ..constant import constant, constant_zero
from ..hierarchy import Node
from .round import round_node
from .squarify import squarify


class TreeMap:
    """
    TreeMap layout
    """

    def __init__(self):
        self._tile = squarify
        self._round = False
        self._dx = 1
        self._dy = 1
        self._padding_stack = {0: 0}
        self._padding_inner = constant_zero
        self._padding_top = constant_zero
        self._padding_right = constant_zero
        self._padding_bottom = constant_zero
        self._padding_left = constant_zero

    def __call__(self, root: Node) -> Node:
        """
        Lays out the specified root hierarchy, assigning the following
        properties on root and its descendants:

        * :code:`node.x0` - the left edge of the rectangle
        * :code:`node.y0` - the top edge of the rectangle
        * :code:`node.x1` - the right edge of the rectangle
        * :code:`node.y1` - the bottom edge of the rectangle

        You must call :code:`root.sum` before passing the hierarchy to the treemap
        layout. You probably also want to call root.sort to order the hierarchy
        before computing the layout.

        Parameters
        ----------
        root : Node
            Root node

        Returns
        -------
        Node
            Node organized as treemap
        """
        root.x0 = 0
        root.y0 = 0
        root.x1 = self._dx
        root.y1 = self._dy
        root.each_before(self._position_node)
        self._padding_stack = {0: 0}
        if self._round:
            root.each_before(round_node)
        return root

    def _position_node(self, node: Node):
        p = self._padding_stack[node.depth]
        x0 = node.x0 + p
        y0 = node.y0 + p
        x1 = node.x1 - p
        y1 = node.y1 - p

        if x1 < x0:
            x0 = x1 = (x0 + x1) * 0.5
        if y1 < y0:
            y0 = y1 = (y0 + y1) * 0.5

        node.x0 = x0
        node.y0 = y0
        node.x1 = x1
        node.y1 = y1

        if node.children:
            p = self._padding_stack[node.depth + 1] = self._padding_inner(node) * 0.5
            x0 += self._padding_left(node) - p
            y0 += self._padding_top(node) - p
            x1 -= self._padding_right(node) - p
            y1 -= self._padding_bottom(node) - p
            if x1 < x0:
                x0 = x1 = (x0 + x1) * 0.5
            if y1 < y0:
                y0 = y1 = (y0 + y1) * 0.5
            self._tile(node, x0, y0, x1, y1)

    def set_round(self, round_value: bool) -> TreeMap:
        """
        If :code:`round` is specified, enables or disables rounding according
        to the given boolean and returns this treemap layout.

        Parameters
        ----------
        round_value : bool
            Round value

        Returns
        -------
        TreeMap
            Itself
        """
        self._round = round_value
        return self

    def set_size(self, size: tuple[float, float]) -> TreeMap:
        """
        If :code:`size` is specified, sets this treemap layout's size to the
        specified two-element array of numbers :code:`[width, height]` and
        returns this treemap layout.

        Parameters
        ----------
        size : tuple[float, float]
            Size values

        Returns
        -------
        TreeMap
            Itself
        """
        self._dx = size[0]
        self._dy = size[1]
        return self

    def set_tile(
        self, tile: Callable[[Node, float, float, float, float], None]
    ) -> TreeMap:
        """
        Sets this treemap layout's tile and returns this treemap layout.

        Available treemap tiles:

        * :func:`d3.treemap_binary <detroit.treemap_binary>`
        * :func:`d3.treemap_dice <detroit.treemap_dice>`
        * :func:`d3.treemap_slice <detroit.treemap_slice>`
        * :func:`d3.treemap_slice_dice <detroit.treemap_slice_dice>`
        * :func:`d3.treemap_squarify <detroit.treemap_squarify>`
        * :func:`d3.treemap_resquarify <detroit.treemap_resquarify>`

        Parameters
        ----------
        tile : Callable[[Node, float, float, float, float], None]
            Tile function

        Returns
        -------
        TreeMap
            Itself
        """
        self._tile = required(tile)
        return self

    def set_padding(self, padding: Callable[[Node], float] | float) -> TreeMap:
        """
        Sets the inner and outer padding to the specified number or function
        and returns this treemap layout.

        Parameters
        ----------
        padding : Callable[[Node], float] | float
            Padding function or constant padding value

        Returns
        -------
        TreeMap
            Itself
        """
        return self.set_padding_inner(padding).set_padding_outer(padding)

    def set_padding_inner(
        self, padding_inner: Callable[[Node], float] | float
    ) -> TreeMap:
        """
        Sets the inner padding to the specified number or function and returns
        this treemap layout. If :code:`padding_inner` is a function, it is
        invoked for each node with children, being passed the current node. The
        inner padding is used to separate a node's adjacent children.

        Parameters
        ----------
        padding_inner : Callable[[Node], float] | float
            Inner padding function or constant inner padding value

        Returns
        -------
        TreeMap
            Itself
        """
        if callable(padding_inner):
            self._padding_inner = padding_inner
        else:
            self._padding_inner = constant(padding_inner)
        return self

    def set_padding_outer(
        self, padding_outer: Callable[[Node], float] | float
    ) -> TreeMap:
        """
        Sets the top, right, bottom and left padding to the specified number or
        function and returns this treemap layout.

        Parameters
        ----------
        padding_outer : Callable[[Node], float] | float
            Outer padding function or constant outer padding value

        Returns
        -------
        TreeMap
            Itself
        """
        return (
            self.set_padding_top(padding_outer)
            .set_padding_right(padding_outer)
            .set_padding_bottom(padding_outer)
            .set_padding_left(padding_outer)
        )

    def set_padding_top(self, padding_top: Callable[[Node], float] | float) -> TreeMap:
        """
        Sets the top padding to the specified number or function and returns
        this treemap layout. If :code:`padding_top` is a function, it is
        invoked for each node with children, being passed the current node. The
        top padding is used to separate the top edge of a node from its
        children.

        Parameters
        ----------
        padding_top : Callable[[Node], float] | float
            Top padding function or constant top padding value

        Returns
        -------
        TreeMap
            Itself
        """
        if callable(padding_top):
            self._padding_top = padding_top
        else:
            self._padding_top = constant(padding_top)
        return self

    def set_padding_right(
        self, padding_right: Callable[[Node], float] | float
    ) -> TreeMap:
        """
        Sets the right padding to the specified number or function and returns
        this treemap layout. If :code:`padding_right` is a function, it is
        invoked for each node with children, being passed the current node. The
        right padding is used to separate the right edge of a node from its
        children.

        Parameters
        ----------
        padding_right : Callable[[Node], float] | float
            Right padding function or constant right padding value

        Returns
        -------
        TreeMap
            Itself
        """
        if callable(padding_right):
            self._padding_right = padding_right
        else:
            self._padding_right = constant(padding_right)
        return self

    def set_padding_bottom(
        self, padding_bottom: Callable[[Node], float] | float
    ) -> TreeMap:
        """
        Sets the bottom padding to the specified number or function and returns
        this treemap layout. If :code:`padding_bottom` is a function, it is
        invoked for each node with children, being passed the current node. The
        bottom padding is used to separate the bottom edge of a node from its
        children.

        Parameters
        ----------
        padding_bottom : Callable[[Node], float] | float
            Bottom padding function or constant bottom padding value

        Returns
        -------
        TreeMap
            Itself
        """
        if callable(padding_bottom):
            self._padding_bottom = padding_bottom
        else:
            self._padding_bottom = constant(padding_bottom)
        return self

    def set_padding_left(
        self, padding_left: Callable[[Node], float] | float
    ) -> TreeMap:
        """
        Sets the left padding to the specified number or function and returns
        this treemap layout. If :code:`padding_left` is a function, it is
        invoked for each node with children, being passed the current node. The
        left padding is used to separate the left edge of a node from its
        children.

        Parameters
        ----------
        padding_left : Callable[[Node], float] | float
            Left padding function or constant left padding value

        Returns
        -------
        TreeMap
            Itself
        """
        if callable(padding_left):
            self._padding_left = padding_left
        else:
            self._padding_left = constant(padding_left)
        return self

    def get_round(self) -> bool:
        return self._round

    def get_size(self) -> tuple[float, float]:
        return [self._dx, self._dy]

    def get_tile(self) -> Callable[[Node, float, float, float, float], None]:
        return self._tile

    def get_padding(self) -> Callable[[Node], float] | float:
        return self.get_padding_inner()

    def get_padding_inner(self) -> Callable[[Node], float] | float:
        return self._padding_inner

    def get_padding_outer(self) -> Callable[[Node], float] | float:
        return self.get_padding_top()

    def get_padding_top(self) -> Callable[[Node], float] | float:
        return self._padding_top

    def get_padding_right(self) -> Callable[[Node], float] | float:
        return self._padding_right

    def get_padding_bottom(self) -> Callable[[Node], float] | float:
        return self._padding_bottom

    def get_padding_left(self) -> Callable[[Node], float] | float:
        return self._padding_left


def treemap() -> TreeMap:
    """
    Builds a new treemap layout with default settings.

    Returns
    -------
    TreeMap
        TreeMap object
    """
    return TreeMap()



from __future__ import annotations


EXPONENT = 3


class PolyIn:
    """
    Polynomial easing; raises t to the specified exponent. If the
    :code:`exponent` is not specified, it defaults to :code:`3`, equivalent to
    :func:`d3.ease_cubic_in <detroit.ease_cubic_in>`.

    Parameters
    ----------
    t : float
        :math:`t \\in [0, 1]`

    Returns
    -------
    float
        Output value of the easing function

    Examples
    --------
    >>> d3.ease_poly_in(0.2)
    0.008000000000000002
    >>> d3.ease_poly_in.exponent(5)(0.2)
    0.0003200000000000001
    """

    def __init__(self, exponent: float):
        self._exponent = exponent

    def __call__(self, t: float) -> float:
        """
        Polynomial easing; raises t to the specified exponent. If the
        :code:`exponent` is not specified, it defaults to :code:`3`, equivalent
        to :func:`d3.ease_cubic_in <detroit.ease_cubic_in>`.

        Parameters
        ----------
        t : float
            :math:`t \\in [0, 1]`

        Returns
        -------
        float
            Output value of the easing function

        Examples
        --------
        >>> d3.ease_poly_in(0.2)
        0.008000000000000002
        >>> d3.ease_poly_in.exponent(5)(0.2)
        0.0003200000000000001
        """
        return pow(t, self._exponent)

    def exponent(self, exponent: float) -> PolyIn:
        """
        Returns a new easing function with :code:`exponent` value as parameter.

        Parameters
        ----------
        exponent : float
            Exponent value

        Returns
        -------
        PolyIn
            New easing function
        """
        return PolyIn(exponent)


ease_poly_in = PolyIn(EXPONENT)


class PolyOut:
    """
    Reverse polynomial easing; equivalent to :code:`1 - d3.ease_poly_in(1 -
    t)`. If the :code:`exponent` is not specified, it defaults to :code:`3`,
    equivalent to :func:`d3.ease_cubic_out <detroit.ease_cubic_out>`.

    Parameters
    ----------
    t : float
        :math:`t \\in [0, 1]`

    Returns
    -------
    float
        Output value of the easing function

    Examples
    --------
    >>> d3.ease_poly_out(0.2)
    0.4879999999999999
    >>> d3.ease_poly_out.exponent(5)(0.2)
    0.6723199999999999
    """

    def __init__(self, exponent: float):
        self._exponent = exponent

    def __call__(self, t: float) -> float:
        """
        Reverse polynomial easing; equivalent to :code:`1 - d3.ease_poly_in(1 -
        t)`. If the :code:`exponent` is not specified, it defaults to
        :code:`3`, equivalent to :func:`d3.ease_cubic_out
        <detroit.ease_cubic_out>`.

        Parameters
        ----------
        t : float
            :math:`t \\in [0, 1]`

        Returns
        -------
        float
            Output value of the easing function

        Examples
        --------
        >>> d3.ease_poly_out(0.2)
        0.4879999999999999
        >>> d3.ease_poly_out.exponent(5)(0.2)
        0.6723199999999999
        """
        return 1 - pow(1 - t, self._exponent)

    def exponent(self, exponent: float) -> PolyOut:
        """
        Returns a new easing function with :code:`exponent` value as parameter.

        Parameters
        ----------
        exponent : float
            Exponent value

        Returns
        -------
        PolyOut
            New easing function
        """
        return PolyOut(exponent)


ease_poly_out = PolyOut(EXPONENT)


class PolyInOut:
    """
    Symmetric polynomial easing; scales :func:`d3.ease_poly_in
    <detroit.ease_poly_in>` :math:`\\forall t \\in [0, 0.5]` and
    :func:`d3.ease_poly_out <detroit.ease_poly_out>` :math:`\\forall t \\in
    [0.5, 1]`. If the :code:`exponent` is not specified, it defaults to
    :code:`3`, equivalent to :func:`d3.ease_cubic <detroit.ease_cubic>`.

    Parameters
    ----------
    t : float
        :math:`t \\in [0, 1]`

    Returns
    -------
    float
        Output value of the easing function

    Examples
    --------
    >>> d3.ease_poly_in_out(0.2)
    0.03200000000000001
    >>> d3.ease_poly_in_out.exponent(5)(0.2)
    0.005120000000000001
    """

    def __init__(self, exponent: float):
        self._exponent = exponent

    def __call__(self, t: float) -> float:
        """
        Symmetric polynomial easing; scales :func:`d3.ease_poly_in
        <detroit.ease_poly_in>` :math:`\\forall t \\in [0, 0.5]` and
        :func:`d3.ease_poly_out <detroit.ease_poly_out>` :math:`\\forall t \\in
        [0.5, 1]`. If the :code:`exponent` is not specified, it defaults to
        :code:`3`, equivalent to :func:`d3.ease_cubic <detroit.ease_cubic>`.

        Parameters
        ----------
        t : float
            :math:`t \\in [0, 1]`

        Returns
        -------
        float
            Output value of the easing function

        Examples
        --------
        >>> d3.ease_poly_in_out(0.2)
        0.03200000000000001
        >>> d3.ease_poly_in_out.exponent(5)(0.2)
        0.005120000000000001
        """
        t *= 2
        if t <= 1:
            return pow(t, self._exponent) * 0.5
        else:
            return (2 - pow(2 - t, self._exponent)) * 0.5

    def exponent(self, exponent: float) -> PolyInOut:
        """
        Returns a new easing function with :code:`exponent` value as parameter.

        Parameters
        ----------
        exponent : float
            Exponent value

        Returns
        -------
        PolyInOut
            New easing function
        """
        return PolyInOut(exponent)


ease_poly_in_out = PolyInOut(EXPONENT)



from collections.abc import Callable

from ..types import U, V
from .object_to_value import interpolate as interpolate_value


def piecewise(
    interpolate: Callable[[U, U], Callable[[float], V]],
    values: list[U] | None = None,
) -> Callable[[float], V]:
    if values is None:
        values = interpolate
        interpolate = interpolate_value

    n = len(values) - 1
    I = [interpolate(values[i], values[i + 1]) for i in range(n)]

    def local_interpolate(t):
        i = max(0, min(n - 1, int(t * n)))
        return I[i](t * n - i)

    return local_interpolate



import math

from ...selection import Selection
from .common import Curve, isvaluable, point_radial


class BumpCurve(Curve):
    def __init__(self, context, x):
        self._context = context
        self._x = x
        self._line = math.nan
        self._x0 = math.nan
        self._y0 = math.nan

    def area_start(self):
        self._line = 0

    def area_end(self):
        self._line = math.nan

    def line_start(self):
        self._point = 0

    def line_end(self):
        if isvaluable(self._line) or (self._line != 0 and self._point == 1):
            self._context.close_path()
        self._line = 1 - self._line

    def point(self, x, y):
        if self._point == 0:
            self._point = 1
            if isvaluable(self._line):
                self._context.line_to(x, y)
            else:
                self._context.move_to(x, y)
        elif self._point == 1:
            self._point = 2
            if self._x:
                self._x0 = (self._x0 + x) / 2
                self._context.bezier_curve_to(self._x0, self._y0, self._x0, y, x, y)
            else:
                self._y0 = (self._y0 + y) / 2
                self._context.bezier_curve_to(self._x0, self._y0, x, self._y0, x, y)
        else:
            if self._x:
                self._x0 = (self._x0 + x) / 2
                self._context.bezier_curve_to(self._x0, self._y0, self._x0, y, x, y)
            else:
                self._y0 = (self._y0 + y) / 2
                self._context.bezier_curve_to(self._x0, self._y0, x, self._y0, x, y)
        self._x0 = x
        self._y0 = y


class BumpRadialCurve(Curve):
    def __init__(self, context):
        self._context = context
        self._x0 = math.nan
        self._y0 = math.nan

    def area_start(self):
        return

    def area_end(self):
        return

    def line_start(self):
        self._point = 0

    def line_end(self):
        return

    def point(self, x, y):
        if self._point == 0:
            self._point = 1
        else:
            p0 = point_radial(self._x0, self._y0)
            self._y0 = (self._y0 + y) / 2
            p1 = point_radial(self._x0, self._y0)
            p2 = point_radial(x, self._y0)
            p3 = point_radial(x, y)
            self._context.move_to(*p0)
            self._context.bezier_curve_to(*p1, *p2, *p3)
        self._x0 = x
        self._y0 = y


def curve_bump_x(context: Selection) -> Curve:
    """
    Produces a Bézier curve between each pair of points, with horizontal
    tangents at each point.

    Parameters
    ----------
    context : Selection
        Context

    Returns
    -------
    Curve
        Curve object

    Examples
    --------

    >>> points = [
    ...     [2, 2],
    ...     [6, 8],
    ...     [10, 10],
    ...     [12, 10],
    ...     [14, 4],
    ...     [20, 4],
    ...     [24, 8],
    ...     [29, 6],
    ...     [32, 4],
    ...     [35, 5],
    ...     [38, 2],
    ... ]
    >>> line = d3.line().curve(d3.curve_bump_x)
    >>> line(points)
    'M2,2C4,2,4,8,6,8C8,8,8,10,10,10C11,10,11,10,12,10C13,10,13,4,14,4C17,4,17,4,20,4C22,4,22,8,24,8C26.500,8,26.500,6,29,6C30.500,6,30.500,4,32,4C33.500,4,33.500,5,35,5C36.500,5,36.500,2,38,2'

    **Result**

    .. image:: ../../figures/light_curve_bump_x.svg
       :align: center
       :class: only-light

    .. image:: ../../figures/dark_curve_bump_x.svg
       :align: center
       :class: only-dark
    """
    return BumpCurve(context, True)


def curve_bump_y(context: Selection) -> Curve:
    """
    Produces a Bézier curve between each pair of points, with vertical tangents
    at each point.

    Parameters
    ----------
    context : Selection
        Context

    Returns
    -------
    Curve
        Curve object

    Examples
    --------

    >>> points = [
    ...     [2, 2],
    ...     [6, 8],
    ...     [10, 10],
    ...     [12, 10],
    ...     [14, 4],
    ...     [20, 4],
    ...     [24, 8],
    ...     [29, 6],
    ...     [32, 4],
    ...     [35, 5],
    ...     [38, 2],
    ... ]
    >>> line = d3.line().curve(d3.curve_bump_y)
    >>> line(points)
    'M2,2C2,5,6,5,6,8C6,9,10,9,10,10C10,10,12,10,12,10C12,7,14,7,14,4C14,4,20,4,20,4C20,6,24,6,24,8C24,7,29,7,29,6C29,5,32,5,32,4C32,4.500,35,4.500,35,5C35,3.500,38,3.500,38,2'

    **Result**

    .. image:: ../../figures/light_curve_bump_y.svg
       :align: center
       :class: only-light

    .. image:: ../../figures/dark_curve_bump_y.svg
       :align: center
       :class: only-dark
    """
    return BumpCurve(context, False)


def curve_bump_radial(context: Selection) -> Curve:
    return BumpRadialCurve(context)



from math import acos, sin

from ...types import Point2D
from ..common import Projection
from .azimuthal import azimuthal_invert, azimuthal_raw
from .projection import geo_projection


class AzimuthalEquidistantRaw:
    def __call__(self, lambda_: float, phi: float) -> Point2D:
        def scale(c: float) -> float:
            c = acos(c)
            return c / sin(c) if c != 0.0 else 0.0

        return azimuthal_raw(scale)(lambda_, phi)

    def invert(self, x: float, y: float) -> Point2D:
        def angle(z: float) -> float:
            return z

        return azimuthal_invert(angle)(x, y)


def geo_azimuthal_equidistant() -> Projection:
    """
    The azimuthal equidistant projection.

    Returns
    -------
    Projection
        Projection object
    """
    return (
        geo_projection(AzimuthalEquidistantRaw())
        .scale(79.4188)
        .set_clip_angle(180 - 1e-3)
    )



from .locale import Locale

locale = Locale({"thousands": ",", "grouping": [3], "currency": ["$", ""]})

locale_format = locale.format
format_prefix = locale.format_prefix


def format_default_locale(definition: dict) -> Locale:
    """
    Equivalent to :code:`d3.format`, except it returns a new
    :code:`Locale` with :code:`locale.format` and
    :code:`locale.format_prefix`

    Parameters
    ----------
    definition : dict
        Definition (todo : add more explanations)

    Returns
    -------
    Locale
        Locale with definition parameters
    """
    return Locale(definition)



from math import sqrt
from random import shuffle
from typing import Any


class Circle:
    def __init__(self, x: float, y: float, r: float):
        self.x = x
        self.y = y
        self.r = r

    @classmethod
    def from_other(cls, circle: Any):
        return cls(
            x=circle.get("x") if isinstance(circle, dict) else circle.x,
            y=circle.get("y") if isinstance(circle, dict) else circle.y,
            r=circle.get("r") if isinstance(circle, dict) else circle.r,
        )


def pack_enclose_random(circles: list[dict[str, float]]) -> Circle:
    """
    Computes the smallest circle that encloses the specified array of circles,
    each of which must have a :code:`circle.r` property specifying the circle's
    radius, and :code:`circle.x` and :code:`circle.y` properties specifying the
    circle's center. The enclosing circle is computed using the
    `Matoušek-Sharir-Welzl algorithm \
    <https://people.inf.ethz.ch/emo/PublFiles/SubexLinProg_ALG16_96.pdf>`_.
    (See also `Apollonius' \
    Problem <https://observablehq.com/@d3/apollonius-problem>`_.)

    Parameters
    ----------
    circles : list[dict[str, float]]
        List of circles defined as dictionaries such as :code:`circle = {"x":
        ..., "y": ..., "r": ...}`

    Returns
    -------
    Circle
        Smallest circle that enclosed the specified array of circles
    """
    i = 0
    circles = [Circle.from_other(circle) for circle in circles]
    shuffle(circles)
    n = len(circles)
    basis = []
    e = None
    while i < n:
        p = circles[i]
        if e and encloses_weak(e, p):
            i += 1
        else:
            basis = extend_basis(basis, p)
            e = enclose_basis(basis)
            i = 0
    return e


def extend_basis(basis: list[Circle], p: Circle):
    if encloses_weak_all(p, basis):
        return [p]

    for i in range(len(basis)):
        if encloses_not(p, basis[i]) and encloses_weak_all(
            enclose_basis_2(basis[i], p), basis
        ):
            return [basis[i], p]

    for i in range(len(basis) - 1):
        for j in range(i + 1, len(basis)):
            if (
                encloses_not(enclose_basis_2(basis[i], basis[j]), p)
                and encloses_not(enclose_basis_2(basis[i], p), basis[j])
                and encloses_not(enclose_basis_2(basis[j], p), basis[i])
                and encloses_weak_all(enclose_basis_3(basis[i], basis[j], p), basis)
            ):
                return [basis[i], basis[j], p]

    raise RuntimeError("This error should not be raised.")


def encloses_not(a: Circle, b: Circle) -> bool:
    dr = a.r - b.r
    dx = b.x - a.x
    dy = b.y - a.y
    return dr < 0 or dr * dr < dx * dx + dy * dy


def encloses_weak(a: Circle, b: Circle) -> bool:
    dr = a.r - b.r + max(a.r, b.r, 1) * 1e-9
    dx = b.x - a.x
    dy = b.y - a.y
    return dr > 0 and dr * dr > dx * dx + dy * dy


def encloses_weak_all(a: Circle, basis: list[Circle]) -> bool:
    for i in range(len(basis)):
        if not encloses_weak(a, basis[i]):
            return False
    return True


def enclose_basis(basis: list[Circle]) -> Circle:
    match len(basis):
        case 1:
            return enclose_basis_1(basis[0])
        case 2:
            return enclose_basis_2(basis[0], basis[1])
        case 3:
            return enclose_basis_3(basis[0], basis[1], basis[2])


def enclose_basis_1(a: Circle) -> Circle:
    return Circle(a.x, a.y, a.r)


def enclose_basis_2(a: Circle, b: Circle) -> Circle:
    x1 = a.x
    y1 = a.y
    r1 = a.r
    x2 = b.x
    y2 = b.y
    r2 = b.r
    x21 = x2 - x1
    y21 = y2 - y1
    r21 = r2 - r1
    length = sqrt(x21 * x21 + y21 * y21)
    return Circle(
        (x1 + x2 + x21 / length * r21) * 0.5,
        (y1 + y2 + y21 / length * r21) * 0.5,
        (length + r1 + r2) * 0.5,
    )


def enclose_basis_3(a: Circle, b: Circle, c: Circle) -> Circle:
    x1 = a.x
    y1 = a.y
    r1 = a.r
    x2 = b.x
    y2 = b.y
    r2 = b.r
    x3 = c.x
    y3 = c.y
    r3 = c.r
    a2 = x1 - x2
    a3 = x1 - x3
    b2 = y1 - y2
    b3 = y1 - y3
    c2 = r2 - r1
    c3 = r3 - r1
    d1 = x1 * x1 + y1 * y1 - r1 * r1
    d2 = d1 - x2 * x2 - y2 * y2 + r2 * r2
    d3 = d1 - x3 * x3 - y3 * y3 + r3 * r3
    ab = a3 * b2 - a2 * b3
    xa = (b2 * d3 - b3 * d2) / (ab * 2) - x1
    xb = (b3 * c2 - b2 * c3) / ab
    ya = (a3 * d2 - a2 * d3) / (ab * 2) - y1
    yb = (a2 * c3 - a3 * c2) / ab
    A = xb * xb + yb * yb - 1
    B = 2 * (r1 + xa * xb + ya * yb)
    C = xa * xa + ya * ya - r1 * r1
    r = -((B + sqrt(B * B - 4 * A * C)) / (2 * A) if abs(A) > 1e-6 else C / B)
    return Circle(x1 + xa + xb * r, y1 + ya + yb * r, r)



from collections.abc import Callable
from math import cos, degrees, radians, sin

from ...types import GeoJSON, Point2D, Vec2D
from ..clip import geo_clip_rectangle
from ..common import PolygonStream, Projection
from ..transform import GeoTransformer
from .fit import fit_extent, fit_height, fit_size, fit_width


def transform(projection: Projection) -> GeoTransformer:
    def point(self, x: float, y: float):
        x = projection([x, y])
        self._stream.point(x[0], x[1])

    return GeoTransformer({"point": point})


def identity(x: PolygonStream) -> PolygonStream:
    return x


class GeoIdentity(Projection):
    def __init__(self):
        self._k = 1
        self._tx = 0
        self._ty = 0
        self._sx = 1
        self._sy = 1
        self._alpha = 0
        self._ca = None
        self._sa = None
        self._x0 = None
        self._y0 = None
        self._x1 = None
        self._y1 = None
        self._kx = 1
        self._ky = 1
        self._transform = transform(self)
        self._postclip = identity
        self._cache = None
        self._cache_stream = None

    def __call__(self, point: Point2D) -> Point2D:
        """
        Returns a new array :code:`[x, y]` (typically in pixels) representing
        the projected point of the given point. The point must be specified as
        a two-element array :code:`[longitude, latitude]` in degrees. May
        return :code:`None` if the specified point has no defined projected position,
        such as when the point is outside the clipping bounds of the
        projection.

        Parameters
        ----------
        point : Point2D
            Point :code:`[longitude, latitude]`

        Returns
        -------
        Point2D
            New projected point :code:`[x, y]`
        """
        x = point[0] * self._kx
        y = point[1] * self._ky
        if self._alpha:
            t = y * self._ca - x * self._sa
            x = x * self._ca + y * self._sa
            y = t
        return [x + self._tx, y + self._ty]

    def invert(self, point: Point2D) -> Point2D:
        """
        Returns a new array :code:`[longitude, latitude]` in degrees
        representing the unprojected point of the given projected point. The
        point must be specified as a two-element array :code:`[x, y]`
        (typically in pixels). May return null if the specified point has no
        defined projected position, such as when the point is outside the
        clipping bounds of the projection.

        This method is only defined on invertible projections.

        Parameters
        ----------
        point : Point2D
            2D point where coordinates are in pixels

        Returns
        -------
        Point2D
            2D point where coordinates are in degrees
        """
        x = point[0] - self._tx
        y = point[1] - self._ty
        if self._alpha:
            t = y * self._ca + x * self._sa
            x = x * self._ca - y * self._sa
            y = t
        return [x / self._kx, y / self._ky]

    def stream(self, stream: PolygonStream) -> PolygonStream:
        """
        Returns a projection stream for the specified output stream. Any input
        geometry is projected before being streamed to the output stream. A
        typical projection involves several geometry transformations: the input
        geometry is first converted to radians, rotated on three axes, clipped
        to the small circle or cut along the antimeridian, and lastly projected
        to the plane with adaptive resampling, scale and translation.

        Parameters
        ----------
        stream : PolygonStream
            Stream object

        Returns
        -------
        PolygonStream
            Modified stream object
        """
        if self._cache and self._cache_stream == stream:
            return self._cache
        self._cache_stream = stream
        self._cache = self._transform(self._postclip(self._cache_stream))
        return self._cache

    def set_preclip(
        self, preclip: Callable[[PolygonStream], PolygonStream]
    ) -> Projection:
        """
        If preclip is specified, sets the projection's spherical clipping to
        the specified function and returns the projection; preclip is a
        function that takes a projection stream and returns a clipped stream.

        Parameters
        ----------
        preclip : Callable[[PolygonStream], PolygonStream]
            Preclip function

        Returns
        -------
        Projection
            Itself
        """
        raise NotImplementedError("Unsupported method for GeoIdentity")

    def set_postclip(
        self, postclip: Callable[[PolygonStream], PolygonStream]
    ) -> Projection:
        """
        If postclip is specified, sets the projection's Cartesian clipping to
        the specified function and returns the projection; postclip is a
        function that takes a projection stream and returns a clipped stream.

        Parameters
        ----------
        postclip : Callable[[PolygonStream], PolygonStream]
            Postclip function

        Returns
        -------
        Projection
            Itself
        """
        self._postclip = postclip
        self._x0 = None
        self._x1 = None
        self._y0 = None
        self._y1 = None
        return self.reset()

    def set_clip_extent(
        self, clip_extent: tuple[Point2D, Point2D] | None = None
    ) -> Projection:
        """
        If extent is specified, sets the projection's viewport clip extent to
        the specified bounds in pixels and returns the projection. The extent
        bounds are specified as an array :math:`[[x_0, y_0], [x_1, y_1]]`,
        where :math:`x_0` is the left-side of the viewport, :math:`y_0` is the
        top, :math:`x_1` is the right and :math:`y_1` is the bottom. If extent
        is :code:`None`, no viewport clipping is performed. If extent is not
        specified, returns the current viewport clip extent which defaults to
        :code:`None`. Viewport clipping is independent of small-circle clipping
        via projection.clipAngle.

        Parameters
        ----------
        clip_extent : tuple[Point2D, Point2D] | None


        Returns
        -------
        Projection
            Itself
        """
        if clip_extent is None:
            self._x0 = None
            self._x1 = None
            self._y0 = None
            self._y1 = None
            self._postclip = identity
        else:
            self._x0 = clip_extent[0][0]
            self._x1 = clip_extent[1][0]
            self._y0 = clip_extent[0][1]
            self._y1 = clip_extent[1][1]
            self._postclip = geo_clip_rectangle(self._x0, self._y0, self._x1, self._y1)
        return self.reset()

    def scale(self, k: float) -> Projection:
        """
        If scale is specified, sets the projection's scale factor to the
        specified value and returns the projection. The scale factor
        corresponds linearly to the distance between projected points; however,
        absolute scale factors are not equivalent across projections.

        Parameters
        ----------
        k : float
            Scale factor

        Returns
        -------
        Projection
            Itself
        """
        self._k = k
        return self.reset()

    def translate(self, translation: Vec2D) -> Projection:
        """
        If translate is specified, sets the projection's translation offset to
        the specified two-element array :code:`[tx, ty]` and returns the
        projection. The translation offset determines the pixel coordinates of
        the projection's center. The default translation offset places
        :math:`[0°,0°]` at the center of a :math:`960 \\times 500` area.

        Parameters
        ----------
        translation : Vec2D
            Translation 2D vector

        Returns
        -------
        Projection
            Itself
        """
        self._tx = translation[0]
        self._ty = translation[1]
        return self.reset()

    def set_center(self, center: Point2D) -> Projection:
        """
        If center is specified, sets the projection's center to the specified
        center, a two-element array of :code:`[longitude, latitude]` in degrees
        and returns the projection.

        Parameters
        ----------
        center : Point2D
            Center point

        Returns
        -------
        Projection
            Itself
        """
        raise NotImplementedError("Unsupported method for GeoIdentity")

    def rotate(
        self, angles: tuple[float, float] | tuple[float, float, float]
    ) -> Projection:
        """
        If angles is specified, sets the projection's three-axis spherical
        rotation to the specified value, which must be a two- or three-element
        array of numbers :math:`[\\lambda, \\phi, \\gamma]` specifying the
        rotation angles in degrees about each spherical axis. (These correspond
        to yaw, pitch and roll).

        Parameters
        ----------
        angles : tuple[float, float] | tuple[float, float, float]
            :math:`[\\lambda, \\phi, \\gamma]` values or :math:`[\\lambda, \\phi]` values

        Returns
        -------
        Projection
            Itself
        """
        raise NotImplementedError("Unsupported method of GeoIdentity")

    def set_angle(self, angle: float) -> Projection:
        """
        If angle is specified, sets the projection's post-projection planar
        rotation angle to the specified angle in degrees and returns the
        projection.

        Parameters
        ----------
        angle : float
            Angle value

        Returns
        -------
        Projection
            Itself
        """
        self._alpha = radians(angle)
        self._sa = sin(self._alpha)
        self._ca = cos(self._alpha)
        return self.reset()

    def set_reflect_x(self, reflect_x: bool) -> Projection:
        """
        If reflect is specified, sets whether or not the x-dimension is
        reflected (negated) in the output. This can be useful to display sky
        and astronomical data with the orb seen from below: right ascension
        (eastern direction) will point to the left when North is pointing up.

        Parameters
        ----------
        reflect_x : bool
            Boolean value

        Returns
        -------
        Projection
            Itself
        """
        self._sx = -1 if reflect_x else 1
        return self.reset()

    def set_reflect_y(self, reflect_y: bool) -> Projection:
        """
        If reflect is specified, sets whether or not the y-dimension is
        reflected (negated) in the output. This is especially useful for
        transforming from standard spatial reference systems, which treat
        positive y as pointing up, to display coordinate systems such as Canvas
        and SVG, which treat positive y as pointing down.

        Parameters
        ----------
        reflect_y : bool
            Boolean value

        Returns
        -------
        Projection
            Itself
        """
        self._sy = -1 if reflect_y else 1
        return self.reset()

    def set_precision(self, precision: float) -> Projection:
        """
        If precision is specified, sets the threshold for the projection's
        adaptive resampling to the specified value in pixels and returns the
        projection. This value corresponds to the Douglas–Peucker distance.

        Parameters
        ----------
        precision : float
            Precision value

        Returns
        -------
        Projection
            Itself
        """
        raise NotImplementedError("Unsupported method of GeoIdentity")

    def fit_extent(self, extent: tuple[Point2D, Point2D], obj: GeoJSON) -> Projection:
        """
        Sets the projection's scale and translate to fit the specified GeoJSON
        object in the center of the given extent. The extent is specified as an
        array :math:`[[x_0, y_0], [x_1, y_1]]`, where :math:`x_0` is the left
        side of the bounding box, :math:`y_0` is the top, :math:`x_1`₁ is the
        right and :math:`y_1` is the bottom. Returns the projection.

        Any clip extent is ignored when determining the new scale and
        translate. The precision used to compute the bounding box of the given
        object is computed at an effective scale of 150.

        Parameters
        ----------
        extent : tuple[Point2D, Point2D]
            Extent values
        obj : GeoJSON
            GeoJSON object

        Returns
        -------
        Projection
            Itself
        """
        return fit_extent(self, extent, obj)

    def fit_size(self, size: Point2D, obj: GeoJSON) -> Projection:
        """
        A convenience method for projection.fit_extent where the top-left
        corner of the extent is [0, 0].

        Parameters
        ----------
        size : Point2D
            Size values
        obj : GeoJSON
            GeoJSON object

        Returns
        -------
        Projection
            Itself
        """
        return fit_size(self, size, obj)

    def fit_width(self, width: float, obj: GeoJSON) -> Projection:
        """
        A convenience method for projection.fit_size where the height is
        automatically chosen from the aspect ratio of object and the given
        constraint on width.

        Parameters
        ----------
        width : float
            Width value
        obj : GeoJSON
            GeoJSON object

        Returns
        -------
        Projection
            Itself
        """
        return fit_width(self, width, obj)

    def fit_height(self, height: float, obj: GeoJSON) -> Projection:
        """
        A convenience method for projection.fit_size where the width is
        automatically chosen from the aspect ratio of object and the given
        constraint on height.

        Parameters
        ----------
        height : float
            Height value
        obj : GeoJSON
            GeoJSON object

        Returns
        -------
        Projection
            Itself
        """
        return fit_height(self, height, obj)

    def reset(self) -> Projection:
        self._kx = self._k * self._sx
        self._ky = self._k * self._sy
        self._cache = None
        self._cache_stream = None
        return self

    def get_postclip(self) -> Callable[[PolygonStream], PolygonStream]:
        return self._postclip

    def get_clip_extent(self) -> tuple[Point2D, Point2D] | None:
        if self._x0 is None:
            return None
        else:
            return [[self._x0, self._y0], [self._x1, self._y1]]

    def get_scale(self) -> float:
        return self._k

    def get_translation(self) -> Vec2D:
        return [self._x, self._y]

    def get_angle(self) -> float:
        return degrees(self._alpha)

    def get_reflect_x(self) -> bool:
        return self._sx < 0

    def get_reflect_y(self) -> bool:
        return self._sy < 0


def geo_identity() -> GeoIdentity:
    return GeoIdentity()



from __future__ import annotations

from collections.abc import Callable
from math import floor, isnan, log, sqrt, ulp
from operator import itemgetter
from typing import Generic

from ..array import argpass, blur2, ticks
from ..types import Accessor, MultiPolygonGeoJSON, Point2D, T
from .constant import constant
from .contours import Contours

default_x = itemgetter(0)
default_y = itemgetter(1)


def default_weight() -> float:
    """
    Default weight function

    Returns
    -------
    float
        Output value :code:`1`
    """
    return 1


class Contour:
    """
    Class used to compute an arbitrary contour without needing to recompute the
    underlying grid.

    Parameters
    ----------
    values : list[float]
        Computed grid values
    contours : Contours
        Contours class
    pow4k : float
        Computed constant value
    transform : Callable[[GeoJSON], GeoJSON]
        Transform function
    """

    def __init__(
        self,
        values: list[float],
        contours: Contours,
        pow4k: float,
        transform: Callable[[MultiPolygonGeoJSON], MultiPolygonGeoJSON],
    ):
        self._values = values
        self._contours = contours
        self._pow4k = pow4k
        self._transform = transform

    def __call__(self, value: float) -> MultiPolygonGeoJSON:
        """
        Compute an arbitrary contour without needing to recompute the
        underlying grid.

        Parameters
        ----------
        value : float
            Threshold value

        Returns
        -------
        GeoJSON
            GeoJSON MultiPolygon geometry object
        """
        c = self._transform(self._contours.contour(self._values, value * self._pow4k))
        c["value"] = value
        return c

    def max(self) -> float:
        """
        Computes the maximum density of the grid.

        Returns
        -------
        float
            Maximum density of the grid
        """
        return max(self._values) / self._pow4k


class Density(Generic[T]):
    """
    Constructs a new density estimator with the default settings.
    """

    def __init__(self):
        self._x = argpass(default_x)
        self._y = argpass(default_y)
        self._weight = argpass(default_weight)
        self._dx = 960
        self._dy = 500
        self._r = 20
        self._k = 2
        self._o = self._r * 3
        self._n = (self._dx + self._o * 2) >> self._k
        self._m = (self._dy + self._o * 2) >> self._k
        self._threshold = constant(20)

    def _grid(self, data: list[T]) -> list[float]:
        """
        Computes density values given data input.

        Parameters
        ----------
        data : list[T]
            Data input

        Returns
        -------
        list[float]
            Grid values
        """
        values = [0.0] * (self._n * self._m)
        pow2k = pow(2, -self._k)
        i = -1

        for d in data:
            xi = (self._x(d, i, data) + self._o) * pow2k
            i += 1
            yi = (self._y(d, i, data) + self._o) * pow2k
            wi = self._weight(d, i, data)
            if (
                wi
                and not isnan(wi)
                and xi >= 0.0
                and xi < self._n
                and yi >= 0.0
                and yi < self._m
            ):
                x0 = floor(xi)
                y0 = floor(yi)
                xt = xi - x0 - 0.5
                yt = yi - y0 - 0.5
                values[x0 + y0 * self._n] += (1 - xt) * (1 - yt) * wi
                values[x0 + 1 + y0 * self._n] += xt * (1 - yt) * wi
                values[x0 + 1 + (y0 + 1) * self._n] += xt * yt * wi
                values[x0 + (y0 + 1) * self._n] += (1 - xt) * yt * wi

        blur2({"data": values, "width": self._n, "height": self._m}, self._r * pow2k)
        return values

    def __call__(self, data: list[T]) -> list[MultiPolygonGeoJSON]:
        """
        Estimates the density contours for the given array of data, returning
        an array of GeoJSON MultiPolygon geometry objects.

        Parameters
        ----------
        data : list[T]
            Data input

        Returns
        -------
        list[GeoJSON]
            List of GeoJSON MultiPolygon geometry objects
        """
        values = self._grid(data)
        tz = self._threshold(values)
        pow4k = pow(2, 2 * self._k)

        if not isinstance(tz, list):
            max_values = max(values)
            tz = ticks(ulp(0.0), max_values / pow4k, tz) if max_values != 0.0 else []

        density = (
            Contours()
            .set_size([self._n, self._m])
            .set_thresholds(list(map(lambda d: d * pow4k, tz)))(values)
        )

        def transform(pair: tuple[int, MultiPolygonGeoJSON]) -> MultiPolygonGeoJSON:
            i, c = pair
            c["value"] = tz[i]
            return self._transform(c)

        return list(map(transform, enumerate(density)))

    def contours(self, data: list[T]) -> Contour:
        """
        Return a :class:`Contour <detroit.contour.density.Contour>` callable
        class that can be used to compute an arbitrary contour on the given
        data without needing to recompute the underlying grid. The returned
        contour class also exposes a :func:`Contour.max
        <detroit.contour.density.Contour.max>` value which represents the
        maximum density of the grid.

        Parameters
        ----------
        data : list[T]
            Data input

        Returns
        -------
        Contour
            Callable class
        """
        contours = Contours().set_size([self._n, self._m])
        values = self._grid(data)
        pow4k = pow(2, 2 * self._k)
        return Contour(values, contours, pow4k, self._transform)

    def _transform(self, geometry: MultiPolygonGeoJSON) -> MultiPolygonGeoJSON:
        """
        Transforms each coordinate of the geometry given density values.

        Parameters
        ----------
        geometry : GeoJSON
            Geometry

        Returns
        -------
        GeoJSON
            Transformed geometry
        """
        for coordinates in geometry["coordinates"]:
            self._transform_polygon(coordinates)
        return geometry

    def _transform_polygon(self, coordinates: list[list[Point2D]]):
        """
        Transforms a polygon of a geometry given density values.

        Parameters
        ----------
        coordinates : list[list[Point2D]]
            Polygon coordinates
        """
        for ring in coordinates:
            self._transform_ring(ring)

    def _transform_ring(self, ring: list[Point2D]):
        """
        Transforms a ring of a polygon given density values.

        Parameters
        ----------
        ring : list[Point2D]
            Ring coordinates
        """
        for point in ring:
            self._transform_point(point)

    def _transform_point(self, coordinates: Point2D):
        """
        Transforms a 2D point given density values.

        Parameters
        ----------
        coordinates : Point2D
            2D point coordinates
        """
        coordinates[0] = coordinates[0] * pow(2, self._k) - self._o
        coordinates[1] = coordinates[1] * pow(2, self._k) - self._o

    def _resize(self) -> Density:
        """
        Recomputes density values from updated attributes and returns itself.

        Returns
        -------
        Density
            Itself
        """
        self._o = self._r * 3
        self._n = int(self._dx + self._o * 2) >> self._k
        self._m = int(self._dy + self._o * 2) >> self._k
        return self

    def x(self, x: Accessor[T, float] | float) -> Density:
        """
        Sets the x coordinate accessor and returns itself.

        Parameters
        ----------
        x : Accessor[T, float] | float
            X accessor or constant value

        Returns
        -------
        Density
            Itself
        """
        if callable(x):
            self._x = argpass(x)
        else:
            self._x = argpass(constant(x))
        return self

    def y(self, y: Accessor[T, float] | float) -> Density:
        """
        Sets the y coordinate accessor and returns itself.

        Parameters
        ----------
        y : Accessor[T, float] | float
            Y accessor or constant value

        Returns
        -------
        Density
            Itself
        """
        if callable(y):
            self._y = argpass(y)
        else:
            self._y = argpass(constant(y))
        return self

    def set_weight(self, weight: Accessor[T, float] | float) -> Density:
        """
        Sets the weight accessor and returns itself.

        Parameters
        ----------
        x : Accessor[T, float] | float
            Weight accessor or constant value

        Returns
        -------
        Density
            Itself
        """
        if callable(weight):
            self._weight = argpass(weight)
        else:
            self._weight = argpass(constant(weight))
        return self

    def set_size(self, size: tuple[float, float]) -> Density:
        """
        Sets the size of the density estimator to the specified bounds and
        returns itself. The size is specified as an array :code:`[width,
        height]` where :code:`width` is the maximum x-value and :code:`height`
        is the maximum y-value.

        Parameters
        ----------
        size : tuple[float, float]
            Size of the density estimator; default value :code:`[960, 500]`

        Returns
        -------
        Density
            Itself
        """
        dx = size[0]
        dy = size[1]
        if dx < 0.0 or dy < 0.0:
            raise ValueError("Invalid size: negative values")
        self._dx = dx
        self._dy = dy
        return self._resize()

    def set_cell_size(self, cell_size: float) -> Density:
        """
        Sets the size of individual cells in the underlying bin grid of the
        positive integer and returns itself.

        Parameters
        ----------
        cell_size : float
            Cell size; default value :code:`4`

        Returns
        -------
        Density
            Itself
        """
        self._k = floor(log(cell_size) / log(2))
        return self._resize()

    def set_thresholds(
        self,
        thresholds: Callable[[list[float], ...], float | list[float]]
        | list[float]
        | float,
    ) -> Density:
        """
        Sets the threshold function to the contour generator and returns
        itself.

        Thresholds are defined as an array of values :math:`[x_0, x_1, ...]`.
        The first generated density contour corresponds to the area where the
        estimated density is greater than or equal to :math:`x_0`; the second
        contour corresponds to the area where the estimated density is greater
        than or equal to :math:`x_1`, and so on. Thus, there is exactly one
        generated MultiPolygon geometry object for each specified threshold
        value; the threshold value is exposed as :code:`geometry["value"]`. The
        first value :math:`x_0` should typically be greater than zero.

        Parameters
        ----------
        thresholds : Callable[[list[float], ...], float | list[float]] | list[float] | float
            Threshold function or array or constant value


        Returns
        -------
        Density
            Itself
        """
        if callable(thresholds):
            self._threshold = thresholds
        else:
            self._threshold = constant(thresholds)
        return self

    def set_bandwidth(self, bandwidth: float) -> Density:
        """
        Sets the bandwidth (the standard deviation) of the Gaussian kernel and
        returns itself.

        Parameters
        ----------
        bandwidth : float
            Bandwidth value

        Returns
        -------
        Density
            Itself
        """
        self._r = (sqrt(4 * bandwidth * bandwidth + 1) - 1) / 2
        return self._resize()

    def get_x(self) -> Accessor[T, float]:
        return self._x

    def get_y(self) -> Accessor[T, float]:
        return self._y

    def get_weight(self) -> Accessor[T, float]:
        return self._weight

    def get_size(self) -> tuple[float, float]:
        return [self._dx, self._dy]

    def get_cell_size(self) -> int:
        return 1 << self._k

    def get_thresholds(self) -> Callable[[list[float], ...], float | list[float]]:
        return self._threshold

    def get_bandwidth(self) -> float:
        return sqrt(self._r * (self._r + 1))



from __future__ import annotations

from collections.abc import Callable

from ..path import Path
from .constant import constant
from .path import WithPath
from .symbols import (
    symbol_asterisk,
    symbol_circle,
    symbol_cross,
    symbol_diamond,
    symbol_diamond2,
    symbol_plus,
    symbol_square,
    symbol_square2,
    symbol_star,
    symbol_times,
    symbol_triangle,
    symbol_triangle2,
    symbol_wye,
)

# Symbols which are designed to be filled
SYMBOLS_FILL = [
    symbol_circle,
    symbol_cross,
    symbol_diamond,
    symbol_square,
    symbol_star,
    symbol_triangle,
    symbol_wye,
]

# Symbols which are designed to be stroked (with a width of 1.5px and round caps).
SYMBOLS_STROKE = [
    symbol_asterisk,
    symbol_circle,
    symbol_diamond2,
    symbol_plus,
    symbol_square2,
    symbol_times,
    symbol_triangle2,
]


class Symbol(WithPath):
    """
    Builds a new symbol generator of the specified symbol type and size.

    Parameters
    ----------
    symbol_type : Callable[[Path, int | float], None] | None
        Symbol type - default :code:`symbol_circle`
    size : Callable[[], int | float] | int | float | None
        Size - Default :code:`64`
    """

    def __init__(
        self,
        symbol_type: Callable[[Path, int | float], None] | None = None,
        size: Callable[[], int | float] | int | float | None = None,
    ):
        super().__init__()
        self._context = None
        self._symbol_type = symbol_type or symbol_circle
        self._size = size if callable(size) else constant(size or 64)

    def __call__(self) -> str | None:
        """
        Generates a symbol. With the default settings,
        invoking the symbol generator produces a circle
        of 64 square pixels.

        Returns
        -------
        str | None
            Symbol path
        """
        if self._context is None:
            buffer = self._path()
            self._context = buffer

        self._symbol_type(self._context, self._size())

        if buffer:
            self._context = None
            return str(buffer)

    def set_symbol_type(
        self, symbol_type: Callable[[Path, int | float], None]
    ) -> Symbol:
        """
        Sets the symbol type and returns this symbol generator.

        Parameters
        ----------
        symbol_type : Callable[[Path, int | float], None]
            Symbol type

        Returns
        -------
        Symbol
            Itself
        """
        self._symbol_type = symbol_type
        return self

    @property
    def symbol_type(self):
        return self._symbol_type

    def set_size(self, size: Callable[[], int | float] | int | float) -> Symbol:
        """
        Sets the size of the symbol and returns this symbol generator.

        Parameters
        ----------
        size : Callable[[], int | float] | int | float
            Size function or constant

        Returns
        -------
        Symbol
            Itself
        """
        self._size = size if callable(size) else constant(size)
        return self

    @property
    def size(self):
        return self._size

    def set_context(self, context: Path) -> Symbol:
        """
        Sets the context and returns this symbol generator.

        Parameters
        ----------
        context : Path
            Context

        Returns
        -------
        Symbol
            Itself
        """
        self._context = context
        return self



from __future__ import annotations

import math
from bisect import bisect
from typing import Any, Generic, overload

from ..types import Number, T
from .init import init_range
from .linear import LinearBase


class ScaleQuantize(LinearBase, Generic[T]):
    def __init__(self):
        self._x0 = 0
        self._x1 = 1
        self._n = 1
        self._domain = [0.5]
        self._range_vals = [0, 1]
        self._unknown = None

    def __call__(self, x: Number | None = None) -> T:
        """
        Given a value in the input domain, returns the corresponding
        value in the output range.

        Parameters
        ----------
        x : Number | None
            Input value

        Returns
        -------
        T
            Output value
        """
        if x is not None and not math.isnan(x):
            return self._range_vals[bisect(self._domain, x, 0, self._n)]
        else:
            return self._unknown

    def rescale(self):
        x0, x1 = self._x0, self._x1
        n = self._n
        self._domain = [((i + 1) * x1 - (i - n) * x0) / (n + 1) for i in range(n)]
        return self

    def set_domain(self, domain: list[Number]) -> ScaleQuantize:
        """
        Sets the scale’s domain to the specified two-element array of numbers.

        Parameters
        ----------
        domain : list[Number]
            Domain

        Returns
        -------
        ScaleQuantize
            Itself
        """
        self._x0, self._x1 = map(float, sorted(domain)[:2])
        return self.rescale()

    def get_domain(self) -> list[Number]:
        return [self._x0, self._x1]

    def set_range(self, range_vals: list[T]) -> ScaleQuantize:
        """
        Sets the scale’s range to the specified array of values

        Parameters
        ----------
        range_vals : list[T]
            Range values

        Returns
        -------
        ScaleQuantize
            Itself
        """
        self._range_vals = list(range_vals)
        self._n = len(self._range_vals) - 1
        return self.rescale()

    def get_range(self) -> list[T]:
        return self._range_vals.copy()

    def invert_extent(self, y: T) -> Number:
        """
        Returns the extent of values in the domain :math:`[x_0, x_1]`
        for the corresponding value in the range: the inverse of quantize.
        This method is useful for interaction, say to determine the value
        in the domain that corresponds to the pixel location under the mouse.

        Parameters
        ----------
        y : T
            Input value

        Returns
        -------
        Number
            Output value
        """
        i = self._range_vals.index(y)
        if i < 0:
            return [math.nan, math.nan]
        elif i < 1:
            return [self._x0, self._domain[0]]
        elif i >= self._n:
            return [self._domain[self._n - 1], self._x1]
        else:
            return [self._domain[i - 1], self._domain[i]]

    def set_unknown(self, unknown: Any) -> ScaleQuantize:
        """
        Sets the scale's unknown value

        Parameters
        ----------
        unknown : Any
            Unknown value

        Returns
        -------
        ScaleQuantize
            Itself
        """
        self._unknown = unknown
        return self

    def get_unknown(self) -> Any:
        return self._unknown

    def get_thresholds(self):
        return self._domain.copy()

    def copy(self):
        return (
            ScaleQuantize()
            .set_domain(self.get_domain())
            .set_range(self.get_range())
            .set_unknown(self.get_unknown())
        )


@overload
def scale_quantize() -> ScaleQuantize[T]: ...


@overload
def scale_quantize(range_vals: list[T]) -> ScaleQuantize[T]: ...


@overload
def scale_quantize(domain: list[Number], range_vals: list[T]) -> ScaleQuantize[T]: ...


def scale_quantize(*args):
    """
    Builds a new quantize scale with the specified domain and range.

    Parameters
    ----------
    domain : list[Number]
        Array of numbers
    range_vals : list[T]
        Array of values

    Returns
    -------
    ScaleQuantize[T]
        Scale object

    Examples
    --------

    >>> scale = d3.scale_quantize([0, 100], d3.SCHEME_BLUES[6])
    >>> for x in range(11):
    ...     x = 10 * x
    ...     print(x, scale(x))
    ...
    ...
    0 #f7fbff
    10 #f7fbff
    20 #deebf7
    30 #c6dbef
    40 #9ecae1
    50 #6baed6
    60 #4292c6
    70 #2171b5
    80 #08519c
    90 #08306b
    100 #08306b
    """
    scale = ScaleQuantize()
    if len(args) == 1:
        return init_range(scale, range_vals=args[0])
    elif len(args) == 2:
        domain, range_vals = args
        return init_range(scale, domain=domain, range_vals=range_vals)
    return init_range(scale)



from .color import color, hsl, rgb
from .cubehelix import cubehelix
from .lab import gray, hcl, lab, lch

__all__ = [
    "color",
    "cubehelix",
    "gray",
    "hcl",
    "hsl",
    "lab",
    "lch",
    "rgb",
]



import math
from collections.abc import Iterable
from datetime import datetime
from itertools import starmap

from ..types import Accessor, T
from .argpass import argpass


def is_valid(value: T) -> bool:
    """
    Check if the value is valid

    Parameters
    ----------
    value : T
        Input value to check

    Returns
    -------
    bool
        :code:`True` if the :code:`value` is defined (different to :code:`None`
        or :code:`nan`)
    """
    return value is not None and (
        isinstance(value, (str, datetime))
        or (isinstance(value, (float, int)) and not math.isnan(value))
    )


def extent(values: Iterable[T], accessor: Accessor[T, T] | None = None) -> tuple[T, T]:
    """
    Returns the minimum and maximum value in
    the given iterable using natural order.

    Parameters
    ----------
    values : Iterable[T]
        Iterator
    accessor : Accessor[T, T] | None
        Accessor function

    Returns
    -------
    tuple[T, T]
        Minimum, maximum

    Examples
    --------

    >>> a = [1, 4, -2, 8]
    >>> d3.extent(a)
    [-2, 8]
    >>> b = [{"a": 8}, {"a": 1}, {"a": 16}]
    >>> d3.extent(b, lambda x: x["a"])
    [1, 16]

    Notes
    -----

    The accessor function can take one, two or three arguments where:

    * the first argument (type :code:`T`) is the value over iteration
    * the second argument (type :code:`int`) is its index
    * the last argument (type :code:`list[T]`) is the input :code:`values` \
    without any modification
    """
    if accessor is None:
        values = list(filter(is_valid, values))
    else:
        accessor = argpass(accessor)

        def access(index: int, value: T) -> T:
            """Access value given the accessor function"""
            return accessor(value, index, values)

        values = list(filter(is_valid, starmap(access, enumerate(values))))

    return [min(values) if values else None, max(values) if values else None]



from __future__ import annotations

from collections.abc import Callable
from math import acos, cos, degrees, pi, radians, sin
from typing import Any

from ..types import GeoJSON, Point2D, Point3D
from .cartesian import cartesian, cartesian_normalize_in_place, spherical
from .common import LineStream
from .constant import constant
from .rotation import RotateRadians

EPISLON = 1e-6
TAU = 2 * pi


def circle_stream(
    stream: LineStream,
    radius: float,
    delta: float,
    direction: float,
    t0: float | None = None,
    t1: float | None = None,
):
    if not delta:
        return

    cos_radius = cos(radius)
    sin_radius = sin(radius)
    step = direction * delta

    if t0 is None:
        t0 = radius + direction * TAU
        t1 = radius - step * 0.5
    else:
        t0 = circle_radius(cos_radius, t0)
        t1 = circle_radius(cos_radius, t1)
        condition = t0 < t1 if direction > 0 else t0 > t1
        if condition:
            t0 += direction * TAU

    t = t0
    while True:
        point = spherical([cos_radius, -sin_radius * cos(t), -sin_radius * sin(t)])
        stream.point(point[0], point[1])
        t -= step
        condition = t < t1 if direction > 0 else t > t1
        if condition:
            break


def circle_radius(cos_radius: float, point: Point3D) -> float:
    point = cartesian(point)
    point[0] -= cos_radius
    cartesian_normalize_in_place(point)
    radius = acos(-point[1])
    radius = -radius if -point[2] < 0 else radius
    return (radius + TAU - EPISLON) % TAU


class GeoCircle:
    """
    Circle generator
    """

    def __init__(self):
        self._center = constant([0, 0])
        self._radius = constant(90)
        self._precision = constant(2)
        self._ring = None
        self._rotate = None
        self._stream = self

    def __call__(self, *args: Any) -> GeoJSON:
        """
        Returns a new GeoJSON geometry object of type "Polygon" approximating a
        circle on the surface of a sphere, with the current center, radius and
        precision. Any arguments are passed to the accessors.


        Parameters
        ----------
        *args : Any
            Arguments passed to the accessors

        Returns
        -------
        GeoJSON
            GeoJSON object
        """
        c = self._center(*args)
        r = radians(self._radius(*args))
        p = radians(self._precision(*args))
        self._ring = []
        self._rotate = RotateRadians(radians(-c[0]), radians(-c[1]), 0).invert
        circle_stream(self._stream, r, p, 1)
        c = {"type": "Polygon", "coordinates": [self._ring]}
        self._ring = None
        self._rotate = None
        return c

    def point(self, x: float, y: float):
        x = self._rotate(x, y)
        self._ring.append(x)
        x[0] = degrees(x[0])
        x[1] = degrees(x[1])

    def set_center(self, center: Callable[..., Point2D] | Point2D) -> GeoCircle:
        """
        If center is specified, sets the circle center to the specified point
        :code:`[longitude, latitude]` in degrees, and returns this circle
        generator. The center may also be specified as a function.

        Parameters
        ----------
        center : Callable[..., Point2D] | Point2D
            Center function or constant value

        Returns
        -------
        GeoCircle
            Itself
        """
        if callable(center):
            self._center = center
        else:
            self._center = constant(center)
        return self

    def set_radius(self, radius: Callable[..., float] | float) -> GeoCircle:
        """
        If radius is specified, sets the circle radius to the specified angle
        in degrees, and returns this circle generator. The radius may also be
        specified as a function.

        Parameters
        ----------
        radius : Callable[..., float] | float
            Radius function or constant value

        Returns
        -------
        GeoCircle
            Itself
        """
        if callable(radius):
            self._radius = radius
        else:
            self._radius = constant(radius)
        return self

    def set_precision(self, precision: Callable[..., float] | float) -> GeoCircle:
        """
        If precision is specified, sets the circle precision to the specified
        angle in degrees, and returns this circle generator. The precision may
        also be specified as a function.

        Parameters
        ----------
        precision : Callable[..., float] | float
            Precision function or constant value

        Returns
        -------
        GeoCircle
            Itself
        """
        if callable(precision):
            self._precision = precision
        else:
            self._precision = constant(precision)
        return self


def geo_circle() -> GeoCircle:
    """
    Returns a new circle generator.

    Returns
    -------
    GeoCircle
        GeoCircle
    """
    return GeoCircle()



from .asterisk import symbol_asterisk
from .circle import symbol_circle
from .cross import symbol_cross
from .diamond import symbol_diamond
from .diamond2 import symbol_diamond2
from .plus import symbol_plus
from .square import symbol_square
from .square2 import symbol_square2
from .star import symbol_star
from .times import symbol_times
from .triangle import symbol_triangle
from .triangle2 import symbol_triangle2
from .wye import symbol_wye

__all__ = [
    "symbol_asterisk",
    "symbol_circle",
    "symbol_cross",
    "symbol_diamond",
    "symbol_diamond2",
    "symbol_plus",
    "symbol_square",
    "symbol_square2",
    "symbol_star",
    "symbol_times",
    "symbol_triangle",
    "symbol_triangle2",
    "symbol_wye",
]



from ..colors import colors
from ..ramp import ramp

scheme = [
    "af8dc3f7f7f77fbf7b",
    "7b3294c2a5cfa6dba0008837",
    "7b3294c2a5cff7f7f7a6dba0008837",
    "762a83af8dc3e7d4e8d9f0d37fbf7b1b7837",
    "762a83af8dc3e7d4e8f7f7f7d9f0d37fbf7b1b7837",
    "762a839970abc2a5cfe7d4e8d9f0d3a6dba05aae611b7837",
    "762a839970abc2a5cfe7d4e8f7f7f7d9f0d3a6dba05aae611b7837",
    "40004b762a839970abc2a5cfe7d4e8d9f0d3a6dba05aae611b783700441b",
    "40004b762a839970abc2a5cfe7d4e8f7f7f7d9f0d3a6dba05aae611b783700441b",
]
SCHEME_PRGN = list(map(colors, scheme))
interpolate_prgn = ramp(SCHEME_PRGN)



from __future__ import annotations

import math
from collections.abc import Callable
from datetime import datetime
from typing import Any, overload

from ..interpolate import interpolate, interpolate_round, piecewise
from ..types import Number, T
from .continuous import identity
from .init import init_interpolator
from .linear import LinearBase
from .log import LogBase, logp, powp, reflect, transform_log, transform_logn
from .pow import transform_pow, transform_sqrt
from .sequential import copy
from .symlog import transform_symlog


class Diverging:
    """
    Diverging scales are similar to linear scales in that they map
    a continuous, numeric input domain to a continuous output range.
    Unlike linear scales, the input domain and output range of a diverging
    scale always have exactly three elements, and the output range is
    typically specified as an interpolator rather than an array of values.
    Diverging scales are typically used for a color encoding; see also
    d3-scale-chromatic. These scales do not expose invert and interpolate
    methods. There are also log, pow, and symlog variants of diverging scales.

    Parameters
    ----------
    t : Callable[[float], float]
        Transform function
    """

    def __init__(self, t: Callable[[float], float]):
        self._transform = t
        self._x0 = 0
        self._x1 = 0.5
        self._x2 = 1
        self._t0 = t(self._x0)
        self._t1 = t(self._x1)
        self._t2 = t(self._x2)
        self._k10 = 0 if self._t0 == self._t1 else 0.5 / (self._t1 - self._t0)
        self._k21 = 0 if self._t1 == self._t2 else 0.5 / (self._t2 - self._t1)
        self._s = -1 if self._t1 < self._t0 else 1
        self._interpolator = identity
        self._clamp = False
        self._unknown = None

    def __call__(self, x: Number) -> float:
        """
        Given a value from the domain, returns the corresponding value from the range.

        Parameters
        ----------
        x : Number
            Input value

        Returns
        -------
        T
            Corresponding value from the range
        """
        if isinstance(x, float) and math.isnan(x):
            return self._unknown
        else:
            x = self._transform(x)
            k = self._k10 if self._s * x < self._s * self._t1 else self._k21
            x = 0.5 + (x - self._t1) * k
            return self._interpolator(max(0, min(1, x)) if self._clamp else x)

    def set_domain(self, domain: list[Number]) -> Diverging:
        """
        Sets the scale's domain to the specified array of numbers

        Parameters
        ----------
        domain : list[Number]
            Domain

        Returns
        -------
        Diverging
            Itself
        """
        self._x0, self._x1, self._x2 = map(float, list(domain)[:3])
        self._t0 = self._transform(self._x0)
        self._t1 = self._transform(self._x1)
        self._t2 = self._transform(self._x2)
        self._k10 = 0 if self._t0 == self._t1 else 0.5 / (self._t1 - self._t0)
        self._k21 = 0 if self._t1 == self._t2 else 0.5 / (self._t2 - self._t1)
        self._s = -1 if self._t1 < self._t0 else 1
        return self

    def get_domain(self) -> list[Number]:
        return [self._x0, self._x1, self._x2]

    def set_clamp(self, clamp: bool) -> Diverging:
        """
        Enables or disables clamping accordingly.

        Parameters
        ----------
        clamp : bool
            Clamp value

        Returns
        -------
        Diverging
            Itself
        """
        self._clamp = bool(clamp)
        return self

    def get_clamp(self) -> bool:
        return self._clamp

    def set_interpolator(self, interpolator: Callable) -> Diverging:
        """
        Sets the scale’s interpolator to the specified function.

        Parameters
        ----------
        interpolator : Callable
            Interpolator function

        Returns
        -------
        Diverging
            Itself
        """
        self._interpolator = interpolator
        return self

    def get_interpolator(self) -> Callable:
        return self._interpolator

    def set_range(self, range_vals: list[T]) -> Diverging:
        """
        The given two-element array is converted
        to an interpolator function using interpolate

        Parameters
        ----------
        range_vals : list[T]
            Two values

        Returns
        -------
        Diverging
            Itself
        """
        self._r0 = float(range_vals[0])
        self._r1 = float(range_vals[1])
        self._r2 = float(range_vals[2])
        self._interpolator = piecewise(interpolate, [self._r0, self._r1, self._r2])
        return self

    def set_range_round(self, range_vals: list[T]) -> Diverging:
        """
        Sets the scale's range to the specified array of values
        and sets scale's interpolator to :code:`interpolate_round`.

        Parameters
        ----------
        range_vals : list[T]
            Range values

        Returns
        -------
        Diverging
            Itself
        """
        self._r0 = float(range_vals[0])
        self._r1 = float(range_vals[1])
        self._r2 = float(range_vals[2])
        self._interpolator = piecewise(
            interpolate_round, [self._r0, self._r1, self._r2]
        )
        return self

    def get_range(self) -> list[T]:
        return [self._interpolator(0), self._interpolator(0.5), self._interpolator(1)]

    def set_unknown(self, unknown: Any) -> Diverging:
        """
        Sets the output value of the scale for undefined
        or NaN input values.

        Parameters
        ----------
        unknown : Any
            Unknown value

        Returns
        -------
        Diverging
            Itself
        """
        self._unknown = unknown
        return self

    def get_unknown(self) -> Any:
        return self._unknown

    def __str__(self) -> str:
        name = self.__class__.__name__
        attrbs = ["domain", "range"]
        attrbs = (f"{a}={getattr(self, f'get_{a}')()}" for a in attrbs)
        attrbs = ", ".join(attrbs)
        return f"{name}({attrbs})"

    def __repr__(self) -> str:
        name = self.__class__.__name__
        addr = id(self)
        return f"<{name} at {hex(addr)}>"


class DivergingLinear(Diverging, LinearBase):
    def __init__(self):
        Diverging.__init__(self, identity)

    def copy(self):
        return copy(self, DivergingLinear())


class DivergingLog(Diverging, LogBase):
    def __init__(self):
        Diverging.__init__(self, identity)
        LogBase.__init__(self)
        self._transform = transform_log
        self.set_domain([0.1, 1, 10])

    def _rescale(self):
        self._logs = logp(self._base)
        self._pows = powp(self._base)
        d = self.get_domain()[0]
        if isinstance(d, datetime):
            d = d.timestamp()
        if d < 0:
            self._logs = reflect(self._logs)
            self._pows = reflect(self._pows)
            self._transform = transform_logn
        else:
            self._transform = transform_log
        return self

    def set_domain(self, domain: list[Number]) -> DivergingLog:
        """
        Sets the scale's domain to the specified array of numbers

        Parameters
        ----------
        domain : list[Number]
            Domain

        Returns
        -------
        DivergingLog
            Itself
        """
        self._x0, self._x1, self._x2 = map(float, list(domain)[:3])
        self._rescale()
        super().set_domain(domain)
        return self

    def copy(self):
        return copy(self, DivergingLog()).base(self.base)


class DivergingSymlog(Diverging):
    def __init__(self, c: Number = 1):
        self._c = c
        super().__init__(transform_symlog(self._c))

    def set_constant(self, c: Number) -> DivergingSymlog:
        """
        Sets the symlog constant to the specified number and returns this scale.

        Parameters
        ----------
        c : Number
            Constant value

        Returns
        -------
        DivergingSymlog
            Itself
        """
        self._c = float(c)
        self._transform = transform_symlog(self._c)
        return self

    def copy(self):
        return copy(self, DivergingSymlog()).set_constant(self.constant)


class DivergingPow(Diverging, LinearBase):
    def __init__(self, t: Callable = identity):
        super().__init__(t)
        self._exponent = 1

    def _rescale(self):
        if self._exponent == 1:
            self._transform = identity
            return self
        elif self._exponent == 0.5:
            self._transform = transform_sqrt
            return self
        else:
            self._transform = transform_pow(self._exponent)
            return self

    def set_exponent(self, exponent: Number) -> DivergingPow:
        """
        Sets the scale's exponent value.

        Parameters
        ----------
        exponent : Number
            Exponent value

        Returns
        -------
        DivergingPow
            Itself
        """
        self._exponent = float(exponent)
        return self._rescale()

    def get_exponent(self):
        return self._exponent

    def copy(self):
        return copy(self, DivergingPow()).set_exponent(self.get_exponent())


@overload
def scale_diverging() -> DivergingLinear: ...


@overload
def scale_diverging(interpolator: Callable) -> DivergingLinear: ...


@overload
def scale_diverging(
    domain: list[Number], interpolator: Callable
) -> DivergingLinear: ...


def scale_diverging(*args):
    """
    Builds a new diverging scale with the specified
    domain and interpolator function or array.

    Parameters
    ----------
    domain : list[Number]
        Array of numbers
    interpolator : Callable
        Interpolator function

    Returns
    -------
    DivergingLinear
        Diverging object

    Examples
    --------

    >>> scale = d3.scale_diverging([-1, 0, 1], d3.interpolate_rdbu)
    >>> steps = 8
    >>> for x in range(steps + 1):
    ...     x = -1 + 2 * x / steps
    ...     print(x, scale(x))
    ...
    ...
    -1.0 rgb(103, 0, 31)
    -0.75 rgb(184, 45, 53)
    -0.5 rgb(228, 130, 104)
    -0.25 rgb(250, 204, 180)
    0.0 rgb(242, 239, 238)
    0.25 rgb(191, 220, 235)
    0.5 rgb(107, 172, 208)
    0.75 rgb(42, 113, 174)
    1.0 rgb(5, 48, 97)
    >>> d3.interpolate_rdbu(0)
    'rgb(103, 0, 31)'
    >>> d3.interpolate_rdbu(1)
    'rgb(5, 48, 97)'
    """
    scale = DivergingLinear()
    if len(args) == 1:
        return init_interpolator(scale, interpolator=args[0])
    elif len(args) == 2:
        domain, interpolator = args
        return init_interpolator(scale, domain=domain, interpolator=interpolator)
    return init_interpolator(scale)


@overload
def scale_diverging_log() -> DivergingLog: ...


@overload
def scale_diverging_log(interpolator: Callable) -> DivergingLog: ...


@overload
def scale_diverging_log(
    domain: list[Number], interpolator: Callable
) -> DivergingLog: ...


def scale_diverging_log(*args):
    """
    Builds a new diverging scale with a logarithmic
    transform, analogous to a log scale.

    Parameters
    ----------
    domain : list[Number]
        Array of numbers
    interpolator : Callable
        Interpolator function

    Returns
    -------
    DivergingLog
        Diverging object

    Examples
    --------

    >>> scale = d3.scale_diverging_log([1, 10, 100], d3.interpolate_rdbu)
    >>> steps = 10
    >>> for x in range(steps + 1):
    ...     x = 2 * x / steps
    ...     x = 10 ** x
    ...     print(x, scale(x))
    ...
    ...
    1.0 rgb(103, 0, 31)
    1.5848931924611136 rgb(172, 32, 47)
    2.51188643150958 rgb(213, 96, 80)
    3.9810717055349722 rgb(240, 162, 133)
    6.309573444801933 rgb(250, 215, 196)
    10.0 rgb(242, 239, 238)
    15.848931924611133 rgb(205, 227, 238)
    25.118864315095795 rgb(143, 194, 220)
    39.810717055349734 rgb(74, 148, 196)
    63.09573444801933 rgb(34, 101, 163)
    100.0 rgb(5, 48, 97)
    >>> d3.interpolate_rdbu(0)
    'rgb(103, 0, 31)'
    >>> d3.interpolate_rdbu(1)
    'rgb(5, 48, 97)'
    """
    scale = DivergingLog()
    if len(args) == 1:
        return init_interpolator(scale, interpolator=args[0])
    elif len(args) == 2:
        domain, interpolator = args
        return init_interpolator(scale, domain=domain, interpolator=interpolator)
    return init_interpolator(scale)


@overload
def scale_diverging_symlog() -> DivergingSymlog: ...


@overload
def scale_diverging_symlog(interpolator: Callable) -> DivergingSymlog: ...


@overload
def scale_diverging_symlog(
    domain: list[Number], interpolator: Callable
) -> DivergingSymlog: ...


def scale_diverging_symlog(*args):
    """
    Builds a new diverging scale with a symmetric
    logarithmic transform, analogous to a symlog scale.

    Parameters
    ----------
    domain : list[Number]
        Array of numbers
    interpolator : Callable
        Interpolator function

    Returns
    -------
    DivergingSymlog
        Diverging object

    Examples
    --------

    >>> scale = d3.scale_diverging_symlog([1, 10, 100], d3.interpolate_rdbu)
    >>> scale = scale.set_constant(2)
    >>> steps = 10
    >>> for x in range(steps + 1):
    ...     x = 2 * x / steps
    ...     x = 10 ** x
    ...     print(x, scale(x))
    ...
    ...
    1.0 rgb(44, 0, 19)
    1.5848931924611136 rgb(79, 0, 27)
    2.51188643150958 rgb(129, 9, 35)
    3.9810717055349722 rgb(181, 41, 51)
    6.309573444801933 rgb(218, 106, 87)
    10.0 rgb(244, 176, 147)
    15.848931924611133 rgb(250, 228, 216)
    25.118864315095795 rgb(228, 237, 242)
    39.810717055349734 rgb(180, 214, 231)
    63.09573444801933 rgb(111, 174, 210)
    100.0 rgb(53, 126, 184)
    >>> d3.interpolate_rdbu(0)
    'rgb(103, 0, 31)'
    >>> d3.interpolate_rdbu(1)
    'rgb(5, 48, 97)'
    """
    scale = DivergingSymlog()
    if len(args) == 1:
        return init_interpolator(scale, interpolator=args[0])
    elif len(args) == 2:
        domain, interpolator = args
        return init_interpolator(scale, domain=domain, interpolator=interpolator)
    return init_interpolator(scale)


@overload
def scale_diverging_pow() -> DivergingPow: ...


@overload
def scale_diverging_pow(interpolator: Callable) -> DivergingPow: ...


@overload
def scale_diverging_pow(
    domain: list[Number], interpolator: Callable
) -> DivergingPow: ...


def scale_diverging_pow(*args):
    """
    Builds a new diverging scale with an exponential
    transform, analogous to a power scale.


    Parameters
    ----------
    domain : list[Number]
        Array of numbers
    interpolator : Callable
        Interpolator function

    Returns
    -------
    DivergingPow
        Diverging object

    Examples
    --------

    >>> from math import sqrt
    >>> scale = d3.scale_diverging_pow([0, 1, 10], d3.interpolate_rdbu)
    >>> scale = scale.set_exponent(2)
    >>> steps = 10
    >>> for x in range(steps + 1):
    ...     x = sqrt(10) * x / steps
    ...     print(x, scale(x))
    ...
    ...
    0.0 rgb(103, 0, 31)
    0.31622776601683794 rgb(140, 13, 37)
    0.6324555320336759 rgb(213, 96, 80)
    0.9486832980505138 rgb(249, 232, 221)
    1.2649110640673518 rgb(233, 239, 242)
    1.5811388300841898 rgb(213, 231, 240)
    1.8973665961010275 rgb(180, 214, 232)
    2.213594362117866 rgb(131, 187, 217)
    2.5298221281347035 rgb(74, 148, 196)
    2.8460498941515415 rgb(36, 103, 166)
    3.1622776601683795 rgb(5, 48, 97)
    >>> d3.interpolate_rdbu(0)
    'rgb(103, 0, 31)'
    >>> d3.interpolate_rdbu(1)
    'rgb(5, 48, 97)'
    """
    scale = DivergingPow()
    if len(args) == 1:
        return init_interpolator(scale, interpolator=args[0])
    elif len(args) == 2:
        domain, interpolator = args
        return init_interpolator(scale, domain=domain, interpolator=interpolator)
    return init_interpolator(scale)


@overload
def scale_diverging_sqrt() -> DivergingPow: ...


@overload
def scale_diverging_sqrt(interpolator: Callable) -> DivergingPow: ...


@overload
def scale_diverging_sqrt(
    domain: list[Number], interpolator: Callable
) -> DivergingPow: ...


def scale_diverging_sqrt(*args):
    """
    Builds a new diverging scale with a square-root
    transform, analogous to a sqrt scale.

    Parameters
    ----------
    domain : list[Number]
        Array of numbers
    interpolator : Callable
        Interpolator function

    Returns
    -------
    DivergingPow
        Diverging object

    Examples
    --------

    >>> scale = d3.scale_diverging_sqrt([0, 1, 10], d3.interpolate_rdbu)
    >>> steps = 10
    >>> for x in range(steps + 1):
    ...     x = 100 * x / steps
    ...     print(x, scale(x))
    ...
    ...
    0.0 rgb(103, 0, 31)
    10.0 rgb(194, 221, 235)
    20.0 rgb(148, 197, 222)
    30.0 rgb(108, 172, 209)
    40.0 rgb(77, 150, 197)
    50.0 rgb(56, 130, 186)
    60.0 rgb(42, 113, 174)
    70.0 rgb(31, 96, 159)
    80.0 rgb(22, 79, 139)
    90.0 rgb(13, 63, 118)
    100.0 rgb(5, 48, 97)
    >>> d3.interpolate_rdbu(0)
    'rgb(103, 0, 31)'
    >>> d3.interpolate_rdbu(1)
    'rgb(5, 48, 97)'
    """
    return scale_diverging_pow(*args).set_exponent(0.5)



from math import asin, cos, sin

from ...types import Point2D


class CylindricalEqualArea:
    def __init__(self, phi0):
        self._cos_phi0 = cos(phi0)

    def __call__(self, lambda_: float, phi: float) -> Point2D:
        return [lambda_ * self._cos_phi0, sin(phi) / self._cos_phi0]

    def invert(self, x: float, y: float) -> Point2D:
        return [x / self._cos_phi0, asin(y * self._cos_phi0)]



from ..interpolate import interpolate_rgb_basis_closed
from ..scale import scale_sequential
from .colors import colors


def ramp_closed(range_values):
    return scale_sequential(interpolate_rgb_basis_closed(colors(range_values))).clamp(
        True
    )



import math
from io import StringIO

from .string_round import string_round

TAU = 2 * math.pi
EPSILON = 1e-6
TAU_EPSILON = TAU - EPSILON


class Path:
    """
    Builds a path serializer

    Parameters
    ----------
    digits : int | None
        Number of digits to round

    Returns
    -------
    Path
        New path serializer
    """

    def __init__(self, digits: int | None = None):
        self.digit = digits or 3
        self._x0 = self._y0 = 0  # start of current subpath
        self._x1 = self._y1 = None  # end of current subpath
        self._string = StringIO()

    def _round(self, *values):
        return (string_round(value, self.digit) for value in values)

    def move_to(self, x: int | float, y: int | float):
        """
        Move to the specified point (x, y).

        Parameters
        ----------
        x : int | float
            x position
        y : int | float
            y position
        """
        self._x0 = self._x1 = x
        self._y0 = self._y1 = y
        x, y = self._round(x, y)
        self._string.write(f"M{x},{y}")

    def close_path(self):
        """
        Ends the current subpath and causes an automatic straight line to be
        drawn from the current point to the initial point of the current
        subpath.
        """
        if self._x1 is not None:
            self._x1, self._y1 = self._x0, self._y0
            self._string.write("Z")

    def line_to(self, x: int | float, y: int | float):
        """
        Draws a straight line from the current point to the specified point (x,
        y).

        Parameters
        ----------
        x : int | float
            x position
        y : int | float
            y position
        """
        self._x1 = x
        self._y1 = y
        x, y = self._round(x, y)
        self._string.write(f"L{x},{y}")

    def quadratic_curve_to(
        self, cpx: int | float, cpy: int | float, x: int | float, y: int | float
    ):
        """
        Draws a quadratic Bézier segment from the current point to the
        specified point (x, y), with the specified control point (cpx, cpy).

        Parameters
        ----------
        cpx : int | float
            Specified control x position
        cpy : int | float
            Specified control y position
        x : int | float
            x position
        y : int | float
            y position
        """
        self._x1 = x
        self._y1 = y
        x, y, cpx, cpy = self._round(x, y, cpx, cpy)
        self._string.write(f"Q{cpx},{cpy},{x},{y}")

    def bezier_curve_to(
        self,
        cpx1: int | float,
        cpy1: int | float,
        cpx2: int | float,
        cpy2: int | float,
        x: int | float,
        y: int | float,
    ):
        """
        Draws a cubic Bézier segment from the current point to the specified
        point (x, y), with the specified control points (cpx1, cpy1) and (cpx2,
        cpy2).

        Parameters
        ----------
        cpx1 : int | float
            Specified control x position 1
        cpy1 : int | float
            Specified control y position 1
        cpx2 : int | float
            Specified control x position 2
        cpy2 : int | float
            Specified control y position 2
        x : int | float
            x position
        y : int | float
            y position
        """
        self._x1 = x
        self._y1 = y
        x, y, cpx1, cpy1, cpx2, cpy2 = self._round(x, y, cpx1, cpy1, cpx2, cpy2)
        self._string.write(f"C{cpx1},{cpy1},{cpx2},{cpy2},{x},{y}")

    def arc_to(
        self,
        x1: int | float,
        y1: int | float,
        x2: int | float,
        y2: int | float,
        r: int | float,
    ):
        """
        Draws a circular arc segment with the specified radius that starts
        tangent to the line between the current point and the specified point
        (x1, y1) and ends tangent to the line between the specified points (x1,
        y1) and (x2, y2). If the first tangent point is not equal to the
        current point, a straight line is drawn between the current point and
        the first tangent point.

        Parameters
        ----------
        x1 : int | float
            Arc x position
        y1 : int | float
            Arc y position
        x2 : int | float
            Tangent x position
        y2 : int | float
            Tangent y position
        r : int | float
            Arc radius
        """
        # Is the radius negative? Error.
        if r < 0:
            raise ValueError(f"negative radius: {r}")

        x0, y0 = self._x1 or 0, self._y1 or 0
        x21, y21 = x2 - x1, y2 - y1
        x01, y01 = x0 - x1, y0 - y1
        l01_2 = x01 * x01 + y01 * y01

        # Is this path empty? Move to (x1,y1).
        if self._x1 is None:
            self._x1 = x1
            self._y1 = y1
            self._string.write(f"M{x1},{y1}")

        # Or, is (x1,y1) coincident with (x0,y0)? Do nothing.
        elif not (l01_2 > EPSILON):
            pass

        # Or, are (x0,y0), (x1,y1) and (x2,y2) collinear?
        # Equivalently, is (x1,y1) coincident with (x2,y2)?
        # Or, is the radius zero? Line to (x1,y1).
        elif not (abs(y01 * x21 - y21 * x01) > EPSILON) or not r:
            self._x1 = int(x1)
            self._y1 = int(y1)
            self._string.write(f"L{x1},{y1}")

        # Otherwise, draw an arc!
        else:
            x20, y20 = x2 - x0, y2 - y0
            l21_2 = x21 * x21 + y21 * y21
            l20_2 = x20 * x20 + y20 * y20
            l21 = math.sqrt(l21_2)
            l01 = math.sqrt(l01_2)
            l = r * math.tan(
                (math.pi - math.acos((l21_2 + l01_2 - l20_2) / (2 * l21 * l01))) / 2
            )
            t01 = l / l01
            t21 = l / l21

            # If the start tangent is not coincident with (x0,y0), line to.
            if abs(t01 - 1) > EPSILON:
                a1, a2 = self._round(x1 + t01 * x01, y1 + t01 * y01)
                self._string.write(f"L{a1},{a2}")

            m1 = int(y01 * x20 > x01 * y20)
            self._x1 = x1 + t21 * x21
            self._y1 = y1 + t21 * y21
            m2, m3 = self._round(self._x1, self._y1)
            self._string.write(f"A{r},{r},0,0,{m1},{m2},{m3}")

    def arc(
        self,
        x: int | float,
        y: int | float,
        r: int | float,
        start_angle: int | float,
        end_angle: int | float,
        ccw: bool = False,
    ):
        """
        Draws a circular arc segment with the specified center (x, y), radius,
        start_angle and end_angle. If anticlockwise is true, the arc is drawn
        in the anticlockwise direction; otherwise, it is drawn in the clockwise
        direction. If the current point is not equal to the starting point of
        the arc, a straight line is drawn from the current point to the start
        of the arc.

        Parameters
        ----------
        x : int | float
            Center x position
        y : int | float
            Center y position
        r : int | float
            Radius
        start_angle : int | float
            Start angle
        end_angle : int | float
            End angle
        ccw : bool
            Clockwise direction
        """
        # Is the radius negative? Error.
        if r < 0:
            raise ValueError(f"negative radius: {r}")

        a0 = start_angle
        a1 = end_angle
        dx = r * math.cos(a0)
        dy = r * math.sin(a0)
        x0 = x + dx
        y0 = y + dy
        cw = 1 ^ bool(ccw)
        da = a0 - a1 if ccw else a1 - a0

        # Is this path empty? Move to (x0, y0).
        if self._x1 is None:
            t1, t2 = self._round(x0, y0)
            self._string.write(f"M{t1},{t2}")

        # Or, is (x0,y0) not coincident with the previous point? Line to (x0,y0).
        elif abs(self._x1 - x0) > EPSILON or abs(self._y1 - y0) > EPSILON:
            t1, t2 = self._round(x0, y0)
            self._string.write(f"L{t1},{t2}")

        # Is this arc empty? We're done.
        if not r:
            return

        # Does the angle go the wrong way ? Flip the direction.
        if da < 0:
            new_da = da % TAU
            da = new_da if new_da else new_da + TAU

        # Is this a complete circle? Draw two arcs to complete the circle.
        if da > TAU_EPSILON:
            self._x1 = x0
            self._y1 = y0
            r, x0, y0, ddx, ddy = self._round(r, x0, y0, x - dx, y - dy)
            self._string.write(
                f"A{r},{r},0,1,{cw},{ddx},{ddy}A{r},{r},0,1,{cw},{x0},{y0}"
            )

        # Is this arc non-empty? Draw an arc!
        elif da > EPSILON:
            self._x1 = x + r * math.cos(a1)
            self._y1 = y + r * math.sin(a1)
            r, x1, y1 = self._round(r, self._x1, self._y1)
            self._string.write(f"A{r},{r},0,{int(da >= math.pi)},{cw},{x1},{y1}")

    def rect(self, x: int | float, y: int | float, w: int | float, h: int | float):
        """
        Creates a new subpath containing just the four points (x, y), (x + w,
        y), (x + w, y + h), (x, y + h), with those four points connected by
        straight lines, and then marks the subpath as closed.

        Parameters
        ----------
        x : int | float
            Rectangle x position
        y : int | float
            Rectangle y position
        w : int | float
            Rectangle width
        h : int | float
            Rectangle height
        """
        self._x0 = self._x1 = x
        self._y0 = self._y1 = y
        x, y, w, h = self._round(x, y, w, h)
        self._string.write(f"M{x},{y}h{w}v{h}h-{w}Z")

    def __str__(self) -> str:
        return self._string.getvalue()



from .exponent import exponent


def precision_fixed(step: int | float) -> int:
    """
    Returns a suggested decimal precision for fixed point
    notation given the specified numeric step value. The
    step represents the minimum absolute difference between
    values that will be formatted.

    Parameters
    ----------
    step : int | float


    Returns
    -------
    int
        Decimal precision

    Examples
    --------

    >>> p = d3.precision_fixed(0.5)
    >>> f = d3.format(f".{p}f")
    >>> f(1)
    '1.0'
    >>> f(1.5)
    '1.5'
    >>> f(2)
    '2.0'
    """
    return max(0, -exponent(abs(step)))



duration_second = 1000
duration_minute = duration_second * 60
duration_hour = duration_minute * 60
duration_day = duration_hour * 24
duration_week = duration_day * 7
duration_month = duration_day * 30
duration_year = duration_day * 365



from collections.abc import Callable

from ..types import T


def interpolate_number_array(
    a: list[T], b: list[T] | None
) -> Callable[[float], list[T]]:
    """
    Returns an interpolator between the two arbitrary values a and b.

    Parameters
    ----------
    a : list[T]
        a value
    b : list[T] | None
        b value

    Returns
    -------
    Callable[[float], list[T]]
        Interpolator function

    Examples
    --------

    >>> interpolator = d3.interpolate_number_array([0, 10, 20], [30, 40, 50])
    >>> interpolator(0)
    [0, 10, 20]
    >>> interpolator(1)
    [30, 40, 50]
    >>> interpolator(0.5)
    [15.0, 25.0, 35.0]
    """
    if b is None:
        b = []
    n = min(len(b), len(a)) if a else 0
    c = list(b)

    def interpolate(t):
        for i in range(n):
            c[i] = a[i] * (1 - t) + b[i] * t
        return c

    return interpolate


def is_number_array(x):
    return isinstance(x, list) and all(isinstance(i, (int, float)) for i in x)



from collections.abc import Callable


def interpolate_round(
    a: int | float | str, b: int | float | str
) -> Callable[[float], int]:
    """
    Returns an interpolator between the two numbers a and b.

    Parameters
    ----------
    a : int | float | str
        a value
    b : int | float | str
        b value

    Returns
    -------
    Callable[[float], int]
        Interpolator function

    Examples
    --------

    >>> interpolator = d3.interpolate_round("1.2", "10.4")
    >>> interpolator(0)
    1
    >>> interpolator(1)
    10
    >>> interpolator(0.5)
    6
    """
    a, b = float(a), float(b)
    return lambda t: round(a * (1 - t) + b * t)



from ..colors import colors
from ..ramp import ramp

scheme = [
    "e9a3c9f7f7f7a1d76a",
    "d01c8bf1b6dab8e1864dac26",
    "d01c8bf1b6daf7f7f7b8e1864dac26",
    "c51b7de9a3c9fde0efe6f5d0a1d76a4d9221",
    "c51b7de9a3c9fde0eff7f7f7e6f5d0a1d76a4d9221",
    "c51b7dde77aef1b6dafde0efe6f5d0b8e1867fbc414d9221",
    "c51b7dde77aef1b6dafde0eff7f7f7e6f5d0b8e1867fbc414d9221",
    "8e0152c51b7dde77aef1b6dafde0efe6f5d0b8e1867fbc414d9221276419",
    "8e0152c51b7dde77aef1b6dafde0eff7f7f7e6f5d0b8e1867fbc414d9221276419",
]
SCHEME_PIYG = list(map(colors, scheme))
interpolate_piyg = ramp(SCHEME_PIYG)



import math

from ...selection import Selection
from .common import Curve, isvaluable


class NaturalCurve(Curve):
    def __init__(self, context):
        self._context = context
        self._line = math.nan
        self._x = []
        self._y = []

    def area_start(self):
        self._line = 0

    def area_end(self):
        self._line = math.nan

    def line_start(self):
        self._x = []
        self._y = []

    def line_end(self):
        x = self._x
        y = self._y
        n = len(x)

        if n != 0:
            if isvaluable(self._line):
                self._context.line_to(x[0], y[0])
            else:
                self._context.move_to(x[0], y[0])

            if n == 2:
                self._context.line_to(x[1], y[1])
            else:
                px = control_points(x)
                py = control_points(y)
                for i1 in range(1, n):
                    i0 = i1 - 1
                    self._context.bezier_curve_to(
                        px[0][i0], py[0][i0], px[1][i0], py[1][i0], x[i1], y[i1]
                    )

        if isvaluable(self._line) or (self._line != 0 and n == 1):
            self._context.close_path()
        self._line = 1 - self._line
        self._x = math.nan
        self._y = math.nan

    def point(self, x, y):
        self._x.append(x)
        self._y.append(y)


def control_points(x):
    n = len(x) - 1
    if n == 0:
        return [[0], [2]]
    a = [None] * n
    b = [None] * n
    r = [None] * n

    a[0] = 0
    b[0] = 2
    r[0] = x[0] + 2 * x[1]
    for i in range(1, n - 1):
        a[i] = 1
        b[i] = 4
        r[i] = 4 * x[i] + 2 * x[i + 1]
    a[n - 1] = 2
    b[n - 1] = 7
    r[n - 1] = 8 * x[n - 1] + x[n]
    for i in range(1, n):
        m = a[i] / b[i - 1]
        b[i] -= m
        r[i] -= m * r[i - 1]
    a[n - 1] = r[n - 1] / b[n - 1]
    for i in range(n - 2, -1, -1):
        a[i] = (r[i] - a[i + 1]) / b[i]
    b[n - 1] = (x[n] + a[n - 1]) / 2
    for i in range(0, n - 1):
        b[i] = 2 * x[i + 1] - a[i + 1]
    return [a, b]


def curve_natural(context: Selection) -> Curve:
    """
    Produces a natural cubic spline with the second derivative of the spline
    set to zero at the endpoints.

    Parameters
    ----------
    context : Selection
        Context

    Returns
    -------
    Curve
        Curve object

    Examples
    --------

    >>> points = [
    ...     [2, 2],
    ...     [6, 8],
    ...     [10, 10],
    ...     [12, 10],
    ...     [14, 4],
    ...     [20, 4],
    ...     [24, 8],
    ...     [29, 6],
    ...     [32, 4],
    ...     [35, 5],
    ...     [38, 2],
    ... ]
    >>> line = d3.line().curve(d3.curve_natural)
    >>> line(points)
    'M2,2C3.293,4.356,4.587,6.711,6,8C7.413,9.289,8.946,9.511,10,10C11.054,10.489,11.630,11.247,12,10C12.370,8.753,12.535,5.503,14,4C15.465,2.497,18.230,2.742,20,4C21.770,5.258,22.546,7.531,24,8C25.454,8.469,27.586,7.136,29,6C30.414,4.864,31.112,3.925,32,4C32.888,4.075,33.968,5.164,35,5C36.032,4.836,37.016,3.418,38,2'

    **Result**

    .. image:: ../../figures/light_curve_natural.svg
       :align: center
       :class: only-light

    .. image:: ../../figures/dark_curve_natural.svg
       :align: center
       :class: only-dark
    """
    return NaturalCurve(context)



import math

from ...selection import Selection
from .basis import BezierTrait
from .common import Curve, isvaluable


class BasisOpenCurve(Curve, BezierTrait):
    def __init__(self, context):
        self._context = context
        self._line = math.nan

    def area_start(self):
        self._line = 0

    def area_end(self):
        self._line = math.nan

    def line_start(self):
        self._x0 = math.nan
        self._y0 = math.nan
        self._x1 = math.nan
        self._y1 = math.nan
        self._point = 0

    def line_end(self):
        if isvaluable(self._line) or (self._line != 0 and self._point == 3):
            self._context.close_path()
        self._line = 1 - self._line

    def point(self, x, y):
        if self._point == 0:
            self._point = 1
        elif self._point == 1:
            self._point = 2
        elif self._point == 2:
            self._point = 3
            x0 = (self._x0 + 4 * self._x1 + x) / 6
            y0 = (self._y0 + 4 * self._y1 + y) / 6
            if isvaluable(self._line):
                self._context.line_to(x0, y0)
            else:
                self._context.move_to(x0, y0)
        elif self._point == 3:
            self._point = 4
            self._bezier_curve_to(x, y)
        else:
            self._bezier_curve_to(x, y)
        self._x0 = self._x1
        self._x1 = x
        self._y0 = self._y1
        self._y1 = y


def curve_basis_open(context: Selection) -> Curve:
    """
    Produces a cubic basis spline using the specified control points. Unlike
    basis, the first and last points are not repeated, and thus the curve
    typically does not intersect these points.

    Parameters
    ----------
    context : Selection
        Context

    Returns
    -------
    Curve
        Curve object

    Examples
    --------

    >>> points = [
    ...     [2, 2],
    ...     [6, 8],
    ...     [10, 10],
    ...     [12, 10],
    ...     [14, 4],
    ...     [20, 4],
    ...     [24, 8],
    ...     [29, 6],
    ...     [32, 4],
    ...     [35, 5],
    ...     [38, 2],
    ... ]
    >>> line = d3.line().curve(d3.curve_basis_open)
    >>> line(points)
    'M6,7.333C7.333,8.667,8.667,9.333,9.667,9.667C10.667,10,11.333,10,12,9C12.667,8,13.333,6,14.667,5C16,4,18,4,19.667,4.667C21.333,5.333,22.667,6.667,24.167,7C25.667,7.333,27.333,6.667,28.667,6C30,5.333,31,4.667,32,4.500C33,4.333,34,4.667,35,4.333'

    **Result**

    .. image:: ../../figures/light_curve_basis_open.svg
       :align: center
       :class: only-light

    .. image:: ../../figures/dark_curve_basis_open.svg
       :align: center
       :class: only-dark
    """
    return BasisOpenCurve(context)



from __future__ import annotations

from collections.abc import Callable
from math import cos, inf, isnan, nan, pi, sin, sqrt

from ..types import Force, SimulationNode
from .lcg import lcg


def x(d: SimulationNode) -> float:
    return d["x"]


def y(d: SimulationNode) -> float:
    return d["y"]


INIIAL_RADIUS = 10
INIIAL_ANGLE = pi * (3 - sqrt(5))


class ForceSimulation:
    def __init__(self, nodes: list[SimulationNode]):
        self._nodes = nodes
        self._alpha = 1
        self._alpha_min = 0.001
        self._alpha_decay = 1 - pow(self._alpha_min, 1 / 300)
        self._alpha_target = 0
        self._velocity_decay = 0.6
        self._forces = {}
        self._random = lcg()
        self._initialize_nodes()

    def tick(self, iterations: float | None = None) -> ForceSimulation:
        """
        Manually steps the simulation by the specified number of iterations,
        and returns the simulation. If :code:`iterations` is not specified, it
        defaults to 1 (single step).

        For each iteration, it increments the current alpha by
        :code:`(alpha_target - alpha) * alpha_decay`; then invokes each
        registered force, passing the new alpha; then decrements each node’s
        velocity by :code:`velocity * velocity_decay`; lastly increments each
        node’s position by velocity.

        This method does not dispatch events; events are only dispatched by the
        internal timer when the simulation is started automatically upon
        creation or by calling simulation.restart. The natural number of ticks
        when the simulation is started is :math:`\\lceil
        \\log(\\alpha_{\\text{min}}) / \\log(1 - \\alpha_{\\text{decay}})
        \\rceil`; by default, this is 300.

        This method can be used in conjunction with simulation.stop to compute
        a static force layout. For large graphs, static layouts should be
        computed in a web worker to avoid freezing the user interface.

        Parameters
        ----------
        iterations : float | None
            Number of iterations

        Returns
        -------
        ForceSimulation
            Itself
        """
        if iterations is None:
            iterations = 1

        for k in range(iterations):
            self._alpha += (self._alpha_target - self._alpha) * self._alpha_decay

            for force in self._forces.values():
                force(self._alpha)

            for node in self._nodes:
                if node.get("fx") is None:
                    node["vx"] *= self._velocity_decay
                    node["x"] += node["vx"]
                else:
                    node["x"] = node["fx"]
                    node["vx"] = 0

                if node.get("fy") is None:
                    node["vy"] *= self._velocity_decay
                    node["y"] += node["vy"]
                else:
                    node["y"] = node["fy"]
                    node["vy"] = 0

        return self

    def _initialize_nodes(self):
        for i, node in enumerate(self._nodes):
            node["index"] = i
            if fx := node.get("fx"):
                node["x"] = fx
            if fy := node.get("fy"):
                node["y"] = fy
            if isnan(node.get("x", nan)) or isnan(node.get("y", nan)):
                radius = INIIAL_RADIUS * sqrt(0.5 + i)
                angle = i * INIIAL_ANGLE
                node["x"] = radius * cos(angle)
                node["y"] = radius * sin(angle)
            if isnan(node.get("vx", nan)) or isnan(node.get("vy", nan)):
                node["vx"] = 0
                node["vy"] = 0

    def _initialize_force(self, force: Force) -> Force:
        if hasattr(force, "initialize"):
            force.initialize(self._nodes, self._random)
        return force

    def find(self, x: float, y: float, radius: float | None = None) -> SimulationNode:
        """
        Returns the node closest to the position :math:`(x,y)` with the given
        search radius.

        Parameters
        ----------
        x : float
            x-coordinate value of the position
        y : float
            y-coordinate value of the position
        radius : float | None
            Radius value

        Returns
        -------
        SimulationNode
            Closest node
        """
        if radius is None:
            radius = inf
        else:
            radius *= radius

        closest = None
        for node in self._nodes:
            dx = x - node["x"]
            dy = y - node["y"]
            d2 = dx * dx + dy * dy
            if d2 < radius:
                closest = node
                radius = d2

        return closest

    def set_nodes(self, nodes: list[SimulationNode]) -> ForceSimulation:
        """
        Sets the simulation's nodes to the specified array of objects,
        initializing their positions and velocities if necessary, and then
        re-initializes any bound forces; returns the simulation.

        Each node must be an dictionary (:code:`dict`). The following
        properties are assigned by the simulation:

        * **index** (:code:`int`) - the node's zero-based index into nodes
        * **x** (:code:`float`) - the node's current x-position
        * **y** (:code:`float`) - the node's current y-position
        * **vx** (:code:`float`) - the node's current x-velocity
        * **vy** (:code:`float`) - the node's current y-velocity
        * **fx** (:code:`float`) - the node's fixed x-position
        * **fy** (:code:`float`) - the node's fixed y-position

        The position :math:`(x,y)` and velocity :math:`(v_x,v_y)` may be
        subsequently modified by forces and by the simulation. If either
        :code:`vx` or :code:`vy` is :code:`nan`, the velocity is initialized to
        :math:`(0,0)`. If either :code:`x` or :code:`y` is :code:`nan`, the
        position is initialized in a `phyllotaxis arrangement
        <https://en.wikipedia.org/wiki/Phyllotaxis>`_, so chosen to ensure a
        deterministic, uniform distribution.

        At the end of each tick, after the application of any forces, a node
        with a defined :code:`node["fx"]` has :code:`node["x"]` reset to this
        value and :code:`node["vx"]` set to zero; likewise, a node with a
        defined :code:`node["fy"]` has :code:`node["y"]` reset to this value
        and :code:`node["vy"]` set to zero. To unfix a node that was previously
        fixed, set :code:`node["fx"]` and :code:`node["fy"]` to :code:`None`,
        or delete these properties.

        If the specified array of nodes is modified, such as when nodes are
        added to or removed from the simulation, this method must be called
        again with the new (or changed) array to notify the simulation and
        bound forces of the change; the simulation does not make a defensive
        copy of the specified array.

        Parameters
        ----------
        nodes : list[SimulationNode]
            Array of nodes

        Returns
        -------
        ForceSimulation
            Itself
        """
        self._nodes = nodes
        self._initialize_nodes()
        for force in self._forces.values():
            self._initialize_force(force)
        return self

    def set_alpha(self, alpha: float) -> ForceSimulation:
        """
        It decreases over time as the simulation "cools down".

        Parameters
        ----------
        alpha : float
            Alpha value

        Returns
        -------
        ForceSimulation
            Itself
        """
        self._alpha = alpha
        return self

    def set_alpha_min(self, alpha_min: float) -> ForceSimulation:
        """
        Sets the minimum alpha to the specified number in the range
        :math:`[0,1]` and returns this simulation.

        The simulation's internal timer stops when the current alpha is less
        than the minimum alpha. The default alpha decay rate of ~0.0228
        corresponds to 300 iterations.

        Parameters
        ----------
        alpha_min : float
            Alpha minimum value

        Returns
        -------
        ForceSimulation
            Itself
        """
        self._alpha_min = alpha_min
        return self

    def set_alpha_decay(self, alpha_decay: float) -> ForceSimulation:
        """
        Sets the alpha decay rate to the specified number in the range
        :math:`[0,1]` and returns this simulation.

        The alpha decay rate determines how quickly the current alpha
        interpolates towards the desired target alpha; since the default target
        alpha is zero, by default this controls how quickly the simulation
        cools. Higher decay rates cause the simulation to stabilize more
        quickly, but risk getting stuck in a local minimum; lower values cause
        the simulation to take longer to run, but typically converge on a
        better layout. To have the simulation run forever at the current alpha,
        set the decay rate to zero; alternatively, set a target alpha greater
        than the minimum alpha.

        Parameters
        ----------
        alpha_decay : float
            Alpha decay value

        Returns
        -------
        ForceSimulation
            Itself
        """
        self._alpha_decay = alpha_decay
        return self

    def set_alpha_target(self, alpha_target: float) -> ForceSimulation:
        """
        Sets the current target alpha to the specified number in the range
        :math:`[0,1]` and returns this simulation.

        Parameters
        ----------
        alpha_target : float
            Alpha target value

        Returns
        -------
        ForceSimulation
            Itself
        """
        self._alpha_target = alpha_target
        return self

    def set_velocity_decay(self, velocity_decay: float) -> ForceSimulation:
        """
        Sets the velocity decay factor to the specified number in the range
        :math:`[0,1]` and returns this simulation.

        The decay factor is akin to atmospheric friction; after the application
        of any forces during a tick, each node's velocity is multiplied by
        :code:`1 - decay`. As with lowering the alpha decay rate, less velocity
        decay may converge on a better solution, but risks numerical
        instabilities and oscillation.

        Parameters
        ----------
        velocity_decay : float
            Velocity decay value

        Returns
        -------
        ForceSimulation
            Itself
        """
        self._velocity_decay = 1 - velocity_decay
        return self

    def set_random_source(
        self, random_source: Callable[[None], float]
    ) -> ForceSimulation:
        """
        Sets the function used to generate random numbers; this should be a
        function that returns a number between 0 (inclusive) and 1 (exclusive).
        If source is not specified, returns this simulation's current random
        source which defaults to a fixed-seed `linear congruential generator
        <https://en.wikipedia.org/wiki/Linear_congruential_generator>`_.

        Parameters
        ----------
        random_source : Callable[[None], float]
            Random source function

        Returns
        -------
        ForceSimulation
            Itself
        """
        self._random = random_source
        for force in self._forces:
            self._initialize_force(force)
        return self

    def set_force(self, name: str, force: Force | None = None) -> ForceSimulation:
        """
        If :code:`force` is not :code:`None`, assigns the :code:`force` for the
        specified :code:`name` and returns this simulation. To remove the
        :code:`force` with the given :code:`name`, pass :code:`None` as the
        :code:`force`.

        Parameters
        ----------
        name : str
            Name of the force
        force : Force | None
            Force value

        Returns
        -------
        ForceSimulation
            Itself
        """
        if force is None:
            self._forces.pop(name)
        else:
            self._forces[name] = self._initialize_force(force)
        return self

    def get_alpha(self) -> float:
        return self._alpha

    def get_alpha_min(self) -> float:
        return self._alpha_min

    def get_alpha_decay(self) -> float:
        return self._alpha_decay

    def get_alpha_target(self) -> float:
        return self._alpha_target

    def get_velocity_decay(self) -> float:
        return 1 - self._velocity_decay

    def get_random_source(self) -> Callable[[None], float]:
        return self._random

    def get_force(self, name: str) -> Force:
        return self._forces.get(name)

    def get_nodes(self) -> list[SimulationNode]:
        return self._nodes


def force_simulation(nodes: list[SimulationNode] | None = None) -> ForceSimulation:
    """
    A force simulation implements a velocity Verlet numerical integrator for
    simulating physical forces on particles (nodes). The simulation assumes a
    constant unit time step :math:`\\Delta t = 1` for each step and a constant
    unit mass :math:`m = 1` for all particles. As a result, a force :math:`F`
    acting on a particle is equivalent to a constant acceleration a over the
    time interval :math:`\\Delta t`, and can be simulated simply by adding to
    the particle's velocity, which is then added to the particle's position.

    Parameters
    ----------
    nodes : list[SimulationNode] | None
        List of nodes

    Returns
    -------
    ForceSimulation
        Simulation object
    """
    if nodes is None:
        nodes = []
    return ForceSimulation(nodes)



from collections.abc import Callable

from ..color import lab as color_lab
from .color import color


def interpolate_lab(a: str, b: str) -> Callable[[float], str]:
    """
    Returns a CIELAB color space interpolator between the two colors a and b.
    The colors a and b need not be in CIELAB; they will be converted to CIELAB
    using d3.lab. The return value of the interpolator is an RGB string.

    Parameters
    ----------
    a : str
        Color string a
    b : str
        Color string b

    Returns
    -------
    Callable[[float], str]
        Interpolator function

    Examples
    --------

    >>> interpolator = d3.interpolate_lab("red", "blue")
    >>> interpolator(0)
    'rgb(255, 0, 0)'
    >>> interpolator(1)
    'rgb(0, 0, 255)'
    >>> interpolator(0.5)
    'rgb(193, 0, 136)'
    """
    start = color_lab(a)
    end = color_lab(b)
    l = color(start.l, end.l)
    a = color(start.a, end.a)
    b = color(start.b, end.b)
    opacity = color(start.opacity, end.opacity)

    def interpolate(t):
        start.l = l(t)
        start.a = a(t)
        start.b = b(t)
        start.opacity = opacity(t)
        return str(start)

    return interpolate



from collections.abc import Callable

from ...selection import Selection
from ...types import Number
from .basis import BasisCurve, curve_basis
from .common import Curve


class BundleCurve(Curve):
    def __init__(self, context, beta):
        self._basis = BasisCurve(context)
        self._beta = beta
        self._x = None
        self._y = None

    def area_start(self):
        return

    def area_end(self):
        return

    def line_start(self):
        self._x = []
        self._y = []
        self._basis.line_start()

    def line_end(self):
        x = self._x
        y = self._y
        j = len(x) - 1

        if j > 0:
            x0 = x[0]
            y0 = y[0]
            dx = x[j] - x0
            dy = y[j] - y0

            i = 0
            while i <= j:
                t = i / j
                self._basis.point(
                    self._beta * x[i] + (1 - self._beta) * (x0 + t * dx),
                    self._beta * y[i] + (1 - self._beta) * (y0 + t * dy),
                )
                i += 1

        self._x = None
        self._y = None
        self._basis.line_end()

    def point(self, x, y):
        self._x.append(x)
        self._y.append(y)


def curve_bundle(
    context_or_beta: Selection | Number,
) -> Callable[[Selection], Curve] | Curve:
    """
    Produces a straightened cubic basis spline using the specified control
    points, with the spline straightened according to the curve's beta. This
    curve is typically used in hierarchical edge bundling to disambiguate
    connections, as proposed by Danny Holten in Hierarchical Edge Bundles:
    Visualization of Adjacency Relations in Hierarchical Data.
    Default value of beta is :code:`0.85`.

    Parameters
    ----------
    context_or_beta : Selection | Number
        Context or beta value in range :math:`[0, 1]` representing the bundle
        strength

    Returns
    -------
    Callable[[Selection], Curve] | Curve
        Curve object or function which makes a curve object with beta value set

    Examples
    --------

    >>> points = [
    ...     [2, 2],
    ...     [6, 8],
    ...     [10, 10],
    ...     [12, 10],
    ...     [14, 4],
    ...     [20, 4],
    ...     [24, 8],
    ...     [29, 6],
    ...     [32, 4],
    ...     [35, 5],
    ...     [38, 2],
    ... ]
    >>> line = d3.line().curve(d3.curve_bundle)
    >>> line(points)
    'M2,2L2.657,2.850C3.313,3.700,4.627,5.400,5.940,6.533C7.253,7.667,8.567,8.233,9.597,8.517C10.627,8.800,11.373,8.800,12.120,7.950C12.867,7.100,13.613,5.400,14.927,4.550C16.240,3.700,18.120,3.700,19.717,4.267C21.313,4.833,22.627,5.967,24.082,6.250C25.537,6.533,27.133,5.967,28.447,5.400C29.760,4.833,30.790,4.267,31.820,4.125C32.850,3.983,33.880,4.267,34.910,3.983C35.940,3.700,36.970,2.850,37.485,2.425L38,2'
    >>> line = d3.line().curve(d3.curve_bundle(0.5))
    >>> line(points)
    'M2,2L2.633,2.500C3.267,3,4.533,4,5.800,4.667C7.067,5.333,8.333,5.667,9.433,5.833C10.533,6,11.467,6,12.400,5.500C13.333,5,14.267,4,15.533,3.500C16.800,3,18.400,3,19.833,3.333C21.267,3.667,22.533,4.333,23.883,4.500C25.233,4.667,26.667,4.333,27.933,4C29.200,3.667,30.300,3.333,31.400,3.250C32.500,3.167,33.600,3.333,34.700,3.167C35.800,3,36.900,2.500,37.450,2.250L38,2'

    **Result**

    .. image:: ../../figures/light_curve_bundle.svg
       :align: center
       :class: only-light

    .. image:: ../../figures/dark_curve_bundle.svg
       :align: center
       :class: only-dark
    """
    if isinstance(context_or_beta, (int, float)):
        beta = context_or_beta
        if beta == 1:
            return curve_basis

        def local_curve(context):
            return BundleCurve(context, beta)

        return local_curve
    context = context_or_beta
    return BundleCurve(context, 0.85)



from ..colors import colors
from ..ramp import ramp

scheme = [
    "ffeda0feb24cf03b20",
    "ffffb2fecc5cfd8d3ce31a1c",
    "ffffb2fecc5cfd8d3cf03b20bd0026",
    "ffffb2fed976feb24cfd8d3cf03b20bd0026",
    "ffffb2fed976feb24cfd8d3cfc4e2ae31a1cb10026",
    "ffffccffeda0fed976feb24cfd8d3cfc4e2ae31a1cb10026",
    "ffffccffeda0fed976feb24cfd8d3cfc4e2ae31a1cbd0026800026",
]
SCHEME_YLORRD = list(map(colors, scheme))
interpolate_ylorrd = ramp(SCHEME_YLORRD)



from ..colors import colors
from ..ramp import ramp

scheme = [
    "fc8d59ffffbf99d594",
    "d7191cfdae61abdda42b83ba",
    "d7191cfdae61ffffbfabdda42b83ba",
    "d53e4ffc8d59fee08be6f59899d5943288bd",
    "d53e4ffc8d59fee08bffffbfe6f59899d5943288bd",
    "d53e4ff46d43fdae61fee08be6f598abdda466c2a53288bd",
    "d53e4ff46d43fdae61fee08bffffbfe6f598abdda466c2a53288bd",
    "9e0142d53e4ff46d43fdae61fee08be6f598abdda466c2a53288bd5e4fa2",
    "9e0142d53e4ff46d43fdae61fee08bffffbfe6f598abdda466c2a53288bd5e4fa2",
]
SCHEME_SPECTRAL = list(map(colors, scheme))
interpolate_spectral = ramp(SCHEME_SPECTRAL)



from .format_decimal import format_decimal_parts


class FormatPrefixAuto:
    def __init__(self):
        self.prefix_exponent = 0

    def __call__(self, x, p):
        d = format_decimal_parts(x, p)
        if not d:
            return str(x)
        coefficient, exponent = d
        self.prefix_exponent = max(-8, min(8, exponent // 3)) * 3
        i = exponent - self.prefix_exponent + 1
        n = len(coefficient)
        if i == n:
            return coefficient
        elif i > n:
            return coefficient + "0" * (i - n)
        elif i > 0:
            return coefficient[:i] + "." + coefficient[i:]
        else:
            return "0." + "0" * (-i) + format_decimal_parts(x, max(0, p + i - 1))[0]


prefix_auto = FormatPrefixAuto()


def format_prefix_auto(x, p):
    return prefix_auto(x, p)



from math import isnan

from ...types import T
from ..series import Series
from .none import offset_none


def offset_silhouette(series: list[Series[T]], order: list[int]):
    """
    Shifts the baseline down such that the center of the streamgraph is always
    at zero.

    Parameters
    ----------
    series : list[Series[T]]
        List of series
    order : list[int]
        List of ordered indices
    """
    n = len(series)
    if n == 0:
        return
    s0 = series[order[0]]
    m = len(s0)
    for j in range(m):
        y = 0.0
        for i in range(n):
            x = series[i][j][1]
            y += 0.0 if isnan(x) else x
        s0[j][0] = -y / 2
        s0[j][1] += s0[j][0]
    offset_none(series, order)



from .diverging import offset_diverging
from .expand import offset_expand
from .none import offset_none
from .silhouette import offset_silhouette
from .wiggle import offset_wiggle

__all__ = [
    "offset_expand",
    "offset_none",
    "offset_diverging",
    "offset_silhouette",
    "offset_wiggle",
]



from ...types import T
from ..series import Series
from .appearance import order_appearance
from .ascending import _sum


def order_inside_out(series: list[Series[T]]) -> list[int]:
    """
    Returns a series order such that the earliest series (according to the
    maximum value) are on the inside and the later series are on the outside.
    This order is recommended for streamgraphs in conjunction with
    :func:`d3.stack_offset_wiggle <stack_offset_wiggle>`.

    Parameters
    ----------
    series : list[Series[T]]
        List of series

    Returns
    -------
    list[int]
        List of ordered indices
    """
    n = len(series)
    sums = list(map(_sum, series))
    order = order_appearance(series)
    top = 0
    bottom = 0
    tops = []
    bottoms = []

    for i in range(n):
        j = order[i]
        if top < bottom:
            top += sums[j]
            tops.append(j)
        else:
            bottom += sums[j]
            bottoms.append(j)

    return bottoms[::-1] + tops



from ..colors import colors
from ..ramp import ramp

scheme = [
    "fde0ddfa9fb5c51b8a",
    "feebe2fbb4b9f768a1ae017e",
    "feebe2fbb4b9f768a1c51b8a7a0177",
    "feebe2fcc5c0fa9fb5f768a1c51b8a7a0177",
    "feebe2fcc5c0fa9fb5f768a1dd3497ae017e7a0177",
    "fff7f3fde0ddfcc5c0fa9fb5f768a1dd3497ae017e7a0177",
    "fff7f3fde0ddfcc5c0fa9fb5f768a1dd3497ae017e7a017749006a",
]
SCHEME_RDPU = list(map(colors, scheme))
interpolate_rdpu = ramp(SCHEME_RDPU)



from math import cos, pi, sin, sqrt

KA = 0.89081309152928522810
KR = sin(pi / 10) / sin(7 * pi / 10)
KX = sin(2 * pi / 10) * KR
KY = -cos(2 * pi / 10) * KR


def symbol_star(context, size):
    r = sqrt(size * KA)
    x = KX * r
    y = KY * r
    context.move_to(0, -r)
    context.line_to(x, y)
    for i in range(1, 5):
        a = 2 * pi * i / 5
        c = cos(a)
        s = sin(a)
        context.line_to(s * r, -c * r)
        context.line_to(c * x - s * y, s * x + c * y)
    context.close_path()
